#include "..\inet\Socket.h"
#include "..\BoardDLLs\brd_generic.h"  // ProcessSystemEvent()
#include "TFEngine_communication.h"
#include "step_TCPIPConnect.h"

StepTCPIPConnect::StepTCPIPConnect(STEPREC& stepRecord) : Step(stepRecord)
{
   host = stepRecord.s1.arr;
   port = stepRecord.s2.arr;
   timeout = stepRecord.s3.arr;
   socketHandleVariable = stepRecord.s5.arr;
   isSSLEnabled = (bool) stepRecord.u1;
}

StepTCPIPConnect::~StepTCPIPConnect(void)
{
}

const char* StepTCPIPConnect::GetStepName(void) const
{
	return("TCP Connect");
}

IV_STATUS StepTCPIPConnect::Execute(PIVRINFO pInfo)
{
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   string parsedHost;
   unsigned short uPort;
   int iTimeout;
   char number[14];
   TCPSocket* socket;
   ACE_INET_Addr hostAddress;
   int ret;
   HANDLE waitObjs[3];
   HANDLE hTimer;
   LARGE_INTEGER liTimeout;
   bool loopAgain;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   if((value = pInfo->variables->ParseParameter("Host address", host.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedHost = value;
      delete[] value;
   }

   if((value = pInfo->variables->ParseParameter("Port", port.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      uPort = (unsigned short) atoi(value);
      delete[] value;
   }

   // --- Parse timeout (in milliseconds)
   if((value = pInfo->variables->ParseParameter("Timeout", timeout.c_str(), "500")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iTimeout = atoi(value);
      delete[] value;

      if((iTimeout < 0) && (iTimeout != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: Timeout '%s' must be >= 0\n",
            pInfo->variables->IsParsedEmpty(timeout.c_str()) ? "@TIMEOUT" : timeout.c_str());
         bParmsValid = false;
      }
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   if(hostAddress.set(uPort, parsedHost.c_str()) != 0)
   {
      pInfo->ivLog->Print("      ERROR: Failed to set host address\n");
      pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(errno));
      return(IVR_FAIL);
   }

   pInfo->ivLog->Print("      Connecting to host '%s'\n", parsedHost.c_str());

   socket = new TCPSocket;
   if(isSSLEnabled)
   {
      socket->EnableSSL();
   }

   if((ret = socket->Connect(hostAddress)) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to connect to host\n");
      pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
      delete socket;
      return(IVR_FAIL);
   }

   // Set timeout timer
   hTimer = CreateWaitableTimer(0, false, 0);
   if(iTimeout != FOREVER)
   {
      liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
      SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
   }

   // Set up events array
   waitObjs[0] = pInfo->messages->GetMessageSignal();  // system event such as shutdown
   waitObjs[1] = socket->GetEventHandle();              // signal from socket
   waitObjs[2] = hTimer;                               // timer

   if(iTimeout != FOREVER)
   {
      liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
      SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
   }

   loopAgain = true;

   // Wait for connection completion or timeout
   while(loopAgain)
   {
      switch(WaitForMultipleObjects(3, waitObjs, false, iTimeout))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            IV_STATUS status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               delete socket;
               return(status);
            }

            break;
         }

         // Connection signalled
         case WAIT_OBJECT_0 + 1:
         {
            // Check status
            if((ret = socket->GetStatus()) != Socket::SUCCESS)
            {
               pInfo->ivLog->Print("      Failed to connect\n");
               pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
               delete socket;
               return(IVR_FAIL);
            }

            pInfo->ivLog->Print("      Connection successful\n");
            loopAgain = false;

            break;
         }

         // Connection timeout
         case WAIT_OBJECT_0 + 2:
         {
            pInfo->ivLog->Print("      ERROR: Timed out trying to connect to host '%s'\n", parsedHost.c_str());
            delete socket;
            return(IVR_TIMEOUT);
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            delete socket;
            return(IVR_FAIL);
         }
      }
   }

   if(isSSLEnabled)
   {
      if((ret = socket->SSLHandshake()) == Socket::SUCCESS)
      {
         pInfo->ivLog->Print("      SSL Handshake successful\n");
      }
      else
      {
         pInfo->ivLog->Print("      ERROR: Unable to complete SSL handshake\n");
         pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
         delete socket;
         return(IVR_FAIL);
      }
   }

   sprintf(number, "%d", socket->get_handle());

   if(!pInfo->variables->Set(socketHandleVariable.c_str(), number))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set Socket Handle variable '%s'\n", socketHandleVariable.c_str());
      delete socket;
      return(IVR_FATAL);
   }

   return(IVR_SUCCESS);
 }
