#include <string>
using std::string;

#pragma hdrstop
#include "..\inet\Socket.h"
#include "..\escutil\iv_str.h"          // HexStr2CharStr
#include "..\BoardDLLs\brd_generic.h"   // ProcessSystemEvent()
#include "TFEngine_communication.h"
#include "step_TCPIPReceive.h"

StepTCPIPReceive::StepTCPIPReceive(STEPREC& stepRecord) : Step(stepRecord)
{
   socketHandle = stepRecord.s1.arr;
   responseVariable = stepRecord.s2.arr;
   timeout = stepRecord.s3.arr;
   minimumLength = stepRecord.s6.arr;
   maximumLength = stepRecord.s4.arr;
   terminationString = stepRecord.s5.arr;
   includeTerminationString = (bool) stepRecord.u1;
   hexString = (bool) stepRecord.u2;
   uppercase = (bool) stepRecord.u3;
}

StepTCPIPReceive::~StepTCPIPReceive(void)
{
}

const char* StepTCPIPReceive::GetStepName(void) const
{
	return("TCP Receive");
}

IV_STATUS StepTCPIPReceive::Execute(PIVRINFO pInfo)
{
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   int iTimeout;
   unsigned int iMinimumLength;
   unsigned int iMaximumLength;
   string parsedTerminationString;
   LARGE_INTEGER liTimeout;
   HANDLE waitObjs[3];
   HANDLE hTimer;
   TCPSocket* socket;
   string returnMessage;
   bool terminatorSought;
   bool terminatorFound = false;
   int ret;
   int bytesWaiting = 0;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   if((value = pInfo->variables->ParseParameter("Socket handle", socketHandle.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      ACE_HANDLE handle = (ACE_HANDLE) atoi(value);
      delete[] value;

      if((socket = TCPSocket::GetSocket(handle)) == NULL)
      {
         pInfo->ivLog->Print("      ERROR: '%d' is not a valid socket handle\n", handle);
         bParmsValid = false;
      }
   }

   if((value = pInfo->variables->ParseParameter("Timeout", timeout.c_str(), "@TIMEOUT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iTimeout = atoi(value);
      delete[] value;

      if((iTimeout < 0) && (iTimeout != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: Silence TIMEOUT '%s' must be >= 0\n",
            pInfo->variables->IsParsedEmpty(timeout.c_str()) ? "@TIMEOUT" : timeout.c_str());
         bParmsValid = false;
      }
   }

   if((value = pInfo->variables->ParseParameter("Minimum response length", minimumLength.c_str(),
      "1", "Empty Minimum response length defaults to no minimum")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iMinimumLength = (hexString ? 2 : 1)*atoi(value);
      delete[] value;

      if(iMinimumLength < 1)
      {
		   pInfo->ivLog->Print("      ERROR: Minimum response length '%s' must be > 0\n", minimumLength.c_str());
         bParmsValid = false;
      }
   }

   if((value = pInfo->variables->ParseParameter("Maximum response length", maximumLength.c_str(),
      "-1", "Empty Maximum response length defaults to no maximum")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iMaximumLength = atoi(value);
      delete[] value;
      if(hexString && (iMaximumLength != FOREVER))
      {
         iMaximumLength *= 2;
      }

      if((iMaximumLength < 1) && (iMaximumLength != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: Maximum response length '%s' must be > 0\n", maximumLength.c_str());
         bParmsValid = false;
      }
   }

   if(bParmsValid && (iMinimumLength > iMaximumLength) && (iMaximumLength != FOREVER))
   {
		pInfo->ivLog->Print("      ERROR: Minimum response length '%s' must be < Maximum response length '%s'\n", minimumLength.c_str(), maximumLength.c_str());
      bParmsValid = false;
   }

   if((value = pInfo->variables->ParseParameter("Termination characters", terminationString.c_str(),
      "", "Empty Termination characters defaults to none")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedTerminationString = value;
      delete[] value;
      terminatorSought = !parsedTerminationString.empty();
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   // Initialise value to empty string
   if(!pInfo->variables->Set(responseVariable.c_str(), ""))
   {
      pInfo->ivLog->Print("      ERROR: Failed to clear response variable '%s'\n", responseVariable.c_str());
      return(IVR_FATAL);
   }

   if((ret = socket->SelectEvents(FD_READ)) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to set socket events to FD_READ\n");
      pInfo->ivLog->Print("         %s\n", Socket::WSAErrorString(ret));
      return(IVR_FAIL);
   }

   // Set timeout timer
   hTimer = CreateWaitableTimer(0, false, 0);
   if(iTimeout != FOREVER)
   {
      liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
      SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
   }

   // Set up events array
   waitObjs[0] = pInfo->messages->GetMessageSignal();  // system event such as shutdown
   waitObjs[1] = hTimer;                               // timer for the timeout value
   waitObjs[2] = socket->GetEventHandle();              // signal from socket

   while((returnMessage.length() < iMinimumLength)
      || (((bytesWaiting > 0) || terminatorSought) && !terminatorFound && ((returnMessage.length() < iMaximumLength) || (iMaximumLength == FOREVER))))
   {
      switch(WaitForMultipleObjects(3, waitObjs, false, INFINITE))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            IV_STATUS status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               CloseHandle(hTimer);
               return(status);
            }

            break;
         }

         // Timed out
         case WAIT_OBJECT_0 + 1:
         {
            pInfo->ivLog->Print("      Timed out waiting for complete response\n");
            // Store as much of the message as we have received in the result variable
            pInfo->variables->Set(responseVariable.c_str(), returnMessage.c_str());
            CloseHandle(hTimer);
            return(IVR_TIMEOUT);
         }

         // Socket signalled
         case WAIT_OBJECT_0 + 2:
         {
            const char* receivedBytes;
            int bytesReceived = 1;

            if((ret = socket->ReceiveBytes(receivedBytes, bytesReceived)) == Socket::SUCCESS)
            {
               if(receivedBytes != NULL)
               {
                  if(hexString)
                  {
                     char temp[3];
                     CharStr2HexStr((const unsigned char*) receivedBytes, temp, 1, uppercase);
                     temp[2] = '\0';
                     returnMessage += (string) temp;
                  }
                  else if(receivedBytes[0] == '\0')
                  {
                     CloseHandle(hTimer);
                     pInfo->ivLog->Print("      WARNING: Receive terminated on NULL character\n");
                     pInfo->variables->Set(responseVariable.c_str(), returnMessage.c_str());
                     return(IVR_SUCCESS);
                  }
                  else
                  {
                     returnMessage += receivedBytes[0];
                  }

                  // Check for the termination string if the termination string is not empty,
                  // and the message is at least as big as the termination string
                  if((returnMessage.length() >= iMinimumLength) && (parsedTerminationString.length() > 0)
                     && (returnMessage.length() >= parsedTerminationString.length()))
                  {
                     // Check the end of the message for the termination string
                     terminatorFound = returnMessage.compare(returnMessage.length() - parsedTerminationString.length(), parsedTerminationString.length(), parsedTerminationString) == 0;

                     // Trim the termination string from the message if requested
                     if(terminatorFound && !includeTerminationString)
                     {
                        returnMessage.erase(returnMessage.length() - parsedTerminationString.length());
                     }
                  }
               }

               socket->GetBytesWaiting(bytesWaiting);
            }
            else if(ret == WSAEWOULDBLOCK)
            {
               // Allow 10 ms for socket to clear
               Sleep(10);
               bytesWaiting = 0;
            }
            else
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               if(ret == WSAENOTCONN)
               {
                  pInfo->ivLog->Print("      ERROR: Socket not connected\n");
               }
               else
               {
                  pInfo->ivLog->Print("      ERROR: Failed to receive\n");
                  pInfo->ivLog->Print("         '%s'\n", errorMessage);
               }

               pInfo->variables->Set(responseVariable.c_str(), errorMessage);
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }

            break;
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            pInfo->variables->Set(responseVariable.c_str(), MsgBuf);
            CloseHandle(hTimer);
            return(IVR_FAIL);
         }
      }
   }

   CloseHandle(hTimer);

   if(!pInfo->variables->Set(responseVariable.c_str(), returnMessage.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set variable to receive message '%s'\n", responseVariable.c_str());
      return(IVR_FATAL);
   }

   return(IVR_SUCCESS);
 }
/*StepTCPIPReceive::Execute*/
