#include <stdlib.h>             // Needed for "atoi".
#include <string.h>             // Needed for "strlen".
#include "step_audiospeak.h"
#include "..\BoardDLLs\brd_generic.h"

StepAudioSpeak::StepAudioSpeak(STEPREC& stepRecord) : Step(stepRecord)
{
   bInt = stepRecord.u2;
   sayFormat = (LanguageModule::SAY_FORMAT) stepRecord.u1;
   text = stepRecord.s2.arr;
}

StepAudioSpeak::~StepAudioSpeak(void)
{
}

const char* StepAudioSpeak::GetStepName(void) const
{
	return("Say");
}

IV_STATUS StepAudioSpeak::Execute(PIVRINFO pInfo)
{
   char* value;
   string parsedText;
   string parsedPath;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

	if(PortRequired(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FATAL);
   }

   if((value = pInfo->variables->ParseParameter("Text", text.c_str())) == NULL)
   {
      return(IVR_FATAL);
   }

   parsedText = value;
   delete[] value;

   switch(sayFormat)
   {
      case LanguageModule::SAY_DATE:
      case LanguageModule::SAY_YEARMONTH:
      case LanguageModule::SAY_MONTHDAY:
      {
         // Only accepts dates of format YYYY/MM/DD
         if(!(
            // YYYY/
            isdigit(parsedText[0]) && isdigit(parsedText[1]) && isdigit(parsedText[2]) && isdigit(parsedText[3]) && (parsedText[4] == '/')
            // MM/
            && isdigit(parsedText[5]) && isdigit(parsedText[6]) && (parsedText[7] == '/')
            // DD
            && isdigit(parsedText[8]) && isdigit(parsedText[9]) && (parsedText[10] == '\0')
            ))
         {
            pInfo->ivLog->Print("      ERROR: Date '%s' must be of format YYYY/MM/DD\n", parsedText.c_str());
            return(IVR_FATAL);
         }
         break;
      }

      case LanguageModule::SAY_24HOUR:
      case LanguageModule::SAY_12HOUR:
      {
         int hour;
         int minute;
         int second;

         // Only accepts time of format HH::MM or HH::MM::SS
         if(!(
            // HH:
            isdigit(parsedText[0]) && isdigit(parsedText[1]) && (parsedText[2] == ':')
            // MM
            && isdigit(parsedText[3]) && isdigit(parsedText[4])
            // end
            && ((parsedText[5] == '\0')
            // or :SS
            || ((parsedText[5] == ':') && isdigit(parsedText[6]) && isdigit(parsedText[7]) && (parsedText[8] == '\0')))
            ))
         {
            pInfo->ivLog->Print("      ERROR: Time '%s' must be of format HH:MM or HH:MM:SS\n", parsedText.c_str());
            return(IVR_FATAL);
         }

         // Check values
         sscanf(parsedText.c_str(), "%d", &hour);
         sscanf(parsedText.substr(3).c_str(), "%d", &minute);

         if(parsedText[5] != '\0')
         {
            sscanf(parsedText.substr(6).c_str(), "%d", &second);
         }
         else
         {
            second = 0;
         }

         if((hour < 0) || (hour > 23))
         {
            pInfo->ivLog->Print("      ERROR: Time '%s' must have HH between 0 and 23\n", parsedText.c_str());
            return(IVR_FATAL);
         }

         if((minute < 0) || (minute > 59))
         {
            pInfo->ivLog->Print("      ERROR: Time '%s' must have MM between 0 and 59\n", parsedText.c_str());
            return(IVR_FATAL);
         }

         if((second <0) || (second > 59))
         {
            pInfo->ivLog->Print("      ERROR: Time '%s' must have SS between 0 and 59\n", parsedText.c_str());
            return(IVR_FATAL);
         }

         break;
      }

      case LanguageModule::SAY_NUMBER:
      case LanguageModule::SAY_MONEY:
      {
         int length = parsedText.length();
         int numberOfCommas = 0;
         int decimalPosition;
         bool invalid = false;
         string::size_type index = parsedText.rfind('.');

         // Find decimal position
         decimalPosition = (index == string::npos) ? length : index;

         // Money can only have 0 or 2 digits after the decimal
         if((sayFormat == LanguageModule::SAY_MONEY) && (decimalPosition != length)
            && (decimalPosition != length - 3))
         {
            pInfo->ivLog->Print("      ERROR: Money '%s' must be an integer or have 2 digits after the decimal\n", parsedText.c_str());
            return(IVR_FATAL);
         }

         // Scan for invalid characters or format
         for(int index = decimalPosition - 1; index >= 0; index--)
         {
            if(parsedText[index] == ',')
            {
               // Improper grouping
               if((decimalPosition - index - numberOfCommas)%3 != 1)
               {
                  pInfo->ivLog->Print("      ERROR: %s '%s' has improperly positioned ','\n", sayFormat == LanguageModule::SAY_NUMBER ? "Number" : "Money", parsedText.c_str());
                  invalid = true;
               }
               numberOfCommas += 1;
            }
            // Multiple decimals
            else if(parsedText[index] == '.')
            {
               pInfo->ivLog->Print("      ERROR: %s '%s' too many decimals\n", sayFormat == LanguageModule::SAY_NUMBER ? "Number" : "Money", parsedText.c_str());
               invalid = true;
            }
            else if(((parsedText[index] == '+') || (parsedText[index] == '-')) && (index != 0))
            {
               // +/- not at beginning
               pInfo->ivLog->Print("      ERROR: %s '%s' can only have +/- at beginning\n", sayFormat == LanguageModule::SAY_NUMBER ? "Number" : "Money", parsedText.c_str());
               invalid = true;
            }
            else if(!isdigit(parsedText[index]))
            {
               pInfo->ivLog->Print("      ERROR: %s '%s' has invalid character '%c'\n", sayFormat == LanguageModule::SAY_NUMBER ? "Number" : "Money", parsedText.c_str(), parsedText[index]);
               invalid = true;
            }
         }

         if(invalid)
         {
            return(IVR_FATAL);
         }

         break;
      }
   }

   pInfo->ivLog->Print("      Speaking '%s'\n", parsedText.c_str());

   // Get language information.
   parsedPath = value = pInfo->variables->FilePath(pInfo->variables->Get("@PROMPT"));
   delete[] value;

   // Alpha-numeric use special function, VoxAlpha...
   return(pInfo->ivBoard->Say(parsedText.c_str(), bInt, parsedPath.c_str(), atoi(pInfo->variables->Get("@VOLUME")), sayFormat));
}
