#include <stdlib.h>               // atoi
#include "step_confjoin.h"
#include "..\BoardDLLs\brd_generic.h"

StepConfJoin::StepConfJoin(STEPREC& stepRecord) : Step(stepRecord)
{
   handle = stepRecord.s1.arr;
   bAutoGainControl = (bool) stepRecord.u1;
   iBeepSetting = stepRecord.u2;
   msgCode1 = stepRecord.s2.arr;
   msgCode2 = stepRecord.s4.arr;
   fileName1 = stepRecord.s3.arr;
   fileName2 = stepRecord.s5.arr;
   format = (SOUND_FORMAT) stepRecord.u3;
}

StepConfJoin::~StepConfJoin(void)
{
}

const char* StepConfJoin::GetStepName(void) const
{
	return("Join Conference");
}

IV_STATUS StepConfJoin::Execute(PIVRINFO pInfo)
{
   // Evaluate all parameters before exiting
   bool         bParmsValid = true;
   char*        value;              // temp vars to parse and validate parameters
   char* scriptCodeFileName;
   string fileName;
   IV_STATUS    status = IVR_SUCCESS;
   
   int          iHandle;            // conference handle after validation
   char*        pathedFileName1 = NULL;    // full path to 1st play file (NULL = nothing to play)
   char*        pathedFileName2 = NULL;    // full path to 2nd play file (NULL = nothing to play)
   SOUND_FORMAT scriptFormat1;      // format of 1st script codes
   SOUND_FORMAT scriptFormat2;      // format of 2nd script codes


   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   // Conferencing will only function on licensed TeleFlow
   if(pInfo->portLicense == UNLICENSED)
   {
      pInfo->ivLog->Print ("      THIS PORT IS UNLICENSED\n");
      return(IVR_NOTSUPPORTED);
   }

   // Parse Conference handle
   if((value = pInfo->variables->ParseParameter("Handle", handle.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iHandle = atoi(value);
      delete[] value;

      if(iHandle <= 0)
      {
      	pInfo->ivLog->Print("      ERROR: Handle '%s' evaluates to invalid value %i\n", handle.c_str(), iHandle);
         bParmsValid = false;
      }
   }
   
   pInfo->ivLog->Print("      Join conference %i\n", iHandle);

   // parse Automatic gain control setting
   // ------------------------------------
   // AGC setting is boolean, no validation necessary
   pInfo->ivLog->Print("      Automatic gain control %s\n", bAutoGainControl ? "ON" : "OFF");

   // parse Introductory beep setting
   // -------------------------------
   switch(iBeepSetting)
   {
      case 0:        // ConfController::BEEP_NONE
      {
         pInfo->ivLog->Print("      No introductory beep\n");
         break;
      }
      case 1:        // ConfController::BEEP_JOIN_ONLY
      {
         pInfo->ivLog->Print("      Introductory beep on join only\n");
         break;
      }
      case 2:        // ConfController::BEEP_JOIN_AND_LEAVE
      {
         pInfo->ivLog->Print("      Introductory beep on join and exit\n");
         break;
      }
      default:
      {
         pInfo->ivLog->Print("      ERROR: Invalid introductory beep setting\n");
         bParmsValid = false;
      }
   }

   // parse Format
   // ------------
   // audio file format will be validated by the board DLL

   // parse 1st play code or file
   // ---------------------------
   scriptFormat1 = TF_BY_EXTENSION;
   pathedFileName2 = NULL;
   //
   if(!msgCode1.empty())
   {
      if((value = pInfo->variables->ParseParameter("1st script code", msgCode1.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else if((status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, scriptFormat1)) != IVR_SUCCESS)
      {
         bParmsValid = false;
      }
      else
      {
         fileName = scriptCodeFileName;
         delete[] scriptCodeFileName;
      }

      delete[] value;
   }
   else if(!fileName1.empty())
   {
      if((value = pInfo->variables->ParseParameter("1st voice file name", fileName1.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         scriptFormat1 = format;
         pathedFileName1 = value;
      }
   }

   // parse 2nd play code or file
   // ---------------------------
   scriptFormat2 = TF_BY_EXTENSION;
   pathedFileName2 = NULL;
   if(!msgCode2.empty())
   {
      if((value = pInfo->variables->ParseParameter("2nd script code", msgCode2.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else if((status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, scriptFormat2)) != IVR_SUCCESS)
      {
         bParmsValid = false;
      }
      else
      {
         fileName = scriptCodeFileName;
         delete[] scriptCodeFileName;
      }

      delete[] value;
   }
   else if(!fileName2.empty())
   {
      if((value = pInfo->variables->ParseParameter("2nd voice file name", fileName2.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         scriptFormat2 = format;
         pathedFileName2 = value;
      }
   }

   if(!bParmsValid)
   {
      delete[] pathedFileName1;
      delete[] pathedFileName2;
      return(IVR_FATAL);
   }

   if((pathedFileName1 != NULL) || (pathedFileName2 != NULL))
   {
      // Connect DSP first, so we can play introductory audio
      status = pInfo->ivBoard->ConfDSP(iHandle, bAutoGainControl, iBeepSetting);
      if(status != IVR_SUCCESS)
      {
         delete[] pathedFileName1;
         delete[] pathedFileName2;
         return(status);
      }

      if(pathedFileName1 != NULL)
      {
         // play 1st file
         pInfo->ivLog->Print("      Playing '%s'\n", pathedFileName1);
         status = pInfo->ivBoard->AudioPlay(pathedFileName1,
                                            false,     // not interruptable
                                            scriptFormat1,
                                            atoi(pInfo->variables->Get("@VOLUME")),
                                            false,     // not fancy
                                            0,         // no play from time
                                            FOREVER);  // no play to time
      }

      if((status == IVR_SUCCESS) && (pathedFileName2 != NULL))
      {
         // play 2nd file
         pInfo->ivLog->Print("      Playing '%s'\n", pathedFileName2);
         status = pInfo->ivBoard->AudioPlay(pathedFileName2,
                                            false,     // not interruptable
                                            scriptFormat2,
                                            atoi(pInfo->variables->Get("@VOLUME")),
                                            false,     // not fancy
                                            0,         // no play from time
                                            FOREVER);  // no play to time
      }

      // Automatically promote play FAIL to FATAL,
      // because a FAIL on conference join means the conference bridge
      // could not accept a new member
      if(status == IVR_FAIL)
      {
         status = IVR_FATAL;
      }
   }

   if(status == IVR_SUCCESS)
   {
      pInfo->ivLog->Print("      Connect new member\n");
      status = pInfo->ivBoard->ConfJoin(iHandle, bAutoGainControl, iBeepSetting);
   }

   delete[] pathedFileName1;
   delete[] pathedFileName2;

   return(status);
}
// StepConfJoin::Execute
