#include <stdlib.h>             // Needed for atoi.

#pragma hdrstop
#include "step_simpullrequest.h"
#include "step_HTTP.h"
#include "..\escutil\security.h"

StepSimpullXMLReq::StepSimpullXMLReq(STEPREC& stepRecord) : Step(stepRecord)
{
   handle = stepRecord.s1.arr;
   timeout = stepRecord.s2.arr;
   XMLServer = stepRecord.s3.arr;
   function = stepRecord.s4.arr;
   params = stepRecord.s6.arr;
   connectTimeout = stepRecord.s5.arr;
   connectRetries = stepRecord.s7.arr;
   scriptCode = stepRecord.s8.arr;
}

StepSimpullXMLReq::~StepSimpullXMLReq(void)
{
}

const char* StepSimpullXMLReq::GetStepName(void) const
{
	return("Simpull Request");
}

// -------------------------------------------------------------------------
//   Name:        StepSimpullXMLReq::Execute                          (Public)
//
//   Purpose:     Combines the functionality of 'StepLoadWebDoc' and
//                'StepXMLLoad' to create a convenience step for the
//                Simpull Voice Personals project.
//
//   Modified:    May 15/02 - TF: Created.
//                Jul 04/02 - TF: Added support for
//                                - optional script play
//                                - MD5 hashing as required by Simpull
//                Sep 17/02 - TF: Added support for tracking values for
//                                all XML requests.
//                Mar  1/03 - TF: Added separate timeout and retry values
//                                for initial connection attempt(s).
//
//   Notes:       - None.
// -------------------------------------------------------------------------
IV_STATUS StepSimpullXMLReq::Execute(PIVRINFO pInfo)
{
   // Programmatic values:
   IV_STATUS       status;
   char*           value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   XML*            pxmlReq;
   string          sRequest;          // STL string to compose user's request
   char            sReqTime[12];      // Time request made
   const char*     pszSecret;         // Simpull's 'Secret' value
   char*           pszHash;           // Hash value to pass with XML request
   StepHTTP        stepHTTP;          // Used to retrieve XML doc
   //
   // For tracking requests:
   char            szTemp[12];        // Temporary string manipulation
   string          sReqID;            // STL string to compose unique request identifier
   const char*     pszMachine;        // Simpull's machine ID
   //
   // User supplied data:
   int             iHandle;           // TeleFlow XML handle number
   string          parsedXMLServer;   // server identifier
   string          parsedFunction;    // Simpull XML API function name
   string          parsedParmTxt;     // Simpull XML API parameters
   const char*     pszPID;            // Simpull partner ID
   const char*     pszUID;            // Simpull user ID
   string          parsedConnectTimeout;
   string          parsedConnectRetries;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   if (pInfo->xmlList == NULL)
   {
      pInfo->ivLog->Print("      ERROR: XML not initialized.\n");
      return IVR_FATAL;
   }

   // --- Retrieve Simpull partner ID
   if(!pInfo->variables->IsVariable("@SIMPULL_PID"))
   {
      pInfo->ivLog->Print("      ERROR: Simpull Partner ID not defined\n");
      pInfo->ivLog->Print("      'Simpull Parameters' step has not been run\n");
      bParmsValid = false;
   }
   else
   {
      pszPID = pInfo->variables->Get("@SIMPULL_PID");
   }

   // --- Retrieve Simpull user ID
   if(!pInfo->variables->IsVariable("@SIMPULL_UID"))
   {
      pInfo->ivLog->Print("      ERROR: Simpull User ID not defined\n");
      pInfo->ivLog->Print("      'Simpull Parameters' step has not been run\n");
      bParmsValid = false;
   }
   else
   {
      pszUID = pInfo->variables->Get("@SIMPULL_UID");
   }

   // --- Retrieve Simpull 'secret' password
   if(!pInfo->variables->IsVariable("@SIMPULL_SECRET"))
   {
      pInfo->ivLog->Print("      ERROR: Simpull Secret not defined\n");
      pInfo->ivLog->Print("      'Simpull Parameters' step has not been run\n");
      bParmsValid = false;
   }
   else
   {
      pszSecret = pInfo->variables->Get("@SIMPULL_SECRET");
   }

   // --- Retrieve Simpull machine ID
   if(!pInfo->variables->IsVariable("@SIMPULL_MACHINE"))
   {
      pInfo->ivLog->Print("      ERROR: Simpull Machine ID not defined.\n");
      pInfo->ivLog->Print("      'Simpull Parameters' step has not been run.\n");
      bParmsValid = false;
   }
   else
   {
      pszMachine = pInfo->variables->Get("@SIMPULL_MACHINE");
   }

   sRequest = "";
   memset(sReqTime, 12, 0);

   // ---
   // --- Evaluate data
   // ---

   // --- Retrieve handle number
   if((value = pInfo->variables->ParseParameter("Handle", handle.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iHandle = atoi(value);
      delete[] value;
   }

   // --- Retrieve XML server identifier
   if((value = pInfo->variables->ParseParameter("Server", XMLServer.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedXMLServer = value;
      delete[] value;
   }

   // --- Retrieve function name
   if((value = pInfo->variables->ParseParameter("Function", function.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedFunction = value;
      delete[] value;
   }
   
   // --- Validate initial connection timeout
   if((value = pInfo->variables->ParseParameter("Initial connection timeout", connectTimeout.c_str(), "@SIMPULL_CONTIMEOUT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedConnectTimeout = value;
      delete[] value;
   }

   // --- Validate initial connection retries
   if((value = pInfo->variables->ParseParameter("Initial connection retries", connectRetries.c_str(), "@SIMPULL_CONRETRIES")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedConnectRetries = value;
      delete[] value;
   }

   // --- Retrieve parameter text
   if((value = pInfo->variables->ParseParameter("Parameters", params.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedParmTxt = value;
      delete[] value;

      // --- Reformat parameters as a single line
      if((value = ComposeParams(parsedParmTxt.c_str(), pInfo)) == NULL)
      {
         pInfo->ivLog->Print("      ERROR: Unable to format parameters.\n");
         bParmsValid = false;
      }
      else
      {
         parsedParmTxt = value;
         delete[] value;
      }
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }
   
   // --- Compose current time as a string
   sprintf(sReqTime, "%i", time(NULL));

   // --- Build unique request ID ('sequence number')
   sprintf(szTemp, "%i", pInfo->iAppID);
   sReqID =  sReqTime;
   sReqID += szTemp;

   // --- Build Simpull hash value
   pszHash = BuildHash (pszSecret, pszUID, pszPID, sReqTime);
   if(pszHash == NULL)
   {
      pInfo->ivLog->Print("      ERROR: Could not create Simpull Hash value\n");
      return IVR_FATAL;
   }

   // ---
   // --- Compose request
   // ---
   sRequest = parsedXMLServer;
   sRequest += "/?arg=";
   sRequest += parsedFunction;
   sRequest += "&pid=";
   sRequest += pszPID;
   sRequest += "&t=";
   sRequest += sReqTime;
   sRequest += "&uid=";
   sRequest += pszUID;
   sRequest += "&h=";
   sRequest += pszHash;
   // for tracking requests:
   sRequest += "&mid=";
   sRequest += pszMachine;
   sRequest += "&rid=";
   sRequest += sReqID;
   // function parameters:
   sRequest += parsedParmTxt;
   delete[] pszHash;

   // ---
   // --- Capture information to @SIMPULL_* variables
   // ---
   if(!pInfo->variables->SetGlobal("@SIMPULL_REQ_ID", sReqID.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Unable to set global variable '@SIMPULL_REQ_ID'.\n");
      // Free memory used
      return(IVR_FATAL);
   }

   if(!pInfo->variables->SetGlobal("@SIMPULL_REQ_FNAME", parsedFunction.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Unable to set global variable '@SIMPULL_REQ_FNAME'.\n");
      // Free memory used
      return(IVR_FATAL);
   }

   if(!pInfo->variables->SetGlobal("@SIMPULL_REQ_STR", sRequest.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Unable to set global variable '@SIMPULL_REQ_STR'.\n");
      return(IVR_FATAL);
   }

   // Blank-out variable to retrieve XML document from Load Web Doc step - TF Mar.5/03
   if(!pInfo->variables->Set("@SIMPULL_XML", ""))
   {
      pInfo->ivLog->Print("      ERROR: Unable to set global variable '@SIMPULL_XML'.\n");
      return(IVR_FATAL);
   }

   // ---
   // --- Create an XML object for this document request
   // ---

   // --- Search for previous XML document open under this handle,
   //     and if found, close it.
   pxmlReq = pInfo->xmlList->FindXML (iHandle);
   if (pxmlReq)
   {
      pInfo->ivLog->Print("      WARNING: Handle %d still open.  Closing handle\n", iHandle);
      delete pxmlReq;
   }

   // --- Create an new XML object
   pxmlReq = new XML (pInfo->ivLog);
   pxmlReq->SetReferenceNo (iHandle);

   // ---
   // --- Retrieve the XML document via HTTP, using Load Web Doc step
   // ---
   pInfo->ivLog->Print("      Retrieve URL = %s\n", sRequest.c_str());

   stepHTTP.SetRequestType(HTTPRequest::HTTP_GET);
   stepHTTP.SetResponseBodyVar("@SIMPULL_XML");    // Variable to receive document
   stepHTTP.SetTimeout(timeout.c_str());                // Timeout waiting for page to come back
   stepHTTP.SetURL(sRequest.c_str());              // Simpull XML API request string
   stepHTTP.SetMessageCode(scriptCode.c_str());         // Optional script to play
   stepHTTP.SetConnectionTimeout(connectTimeout.c_str());
   stepHTTP.SetConnectionRetries(parsedConnectRetries.c_str());

   status = stepHTTP.Execute(pInfo);

   if (status != IVR_SUCCESS)
   {
      // Error code will have been reported by StepLoadWebDoc
      pInfo->ivLog->Print("      Unable to load web document.\n");
      return (status);
   }

   // ---
   // --- Store into the XML object
   // ---
   if(!pInfo->variables->IsVariable("@SIMPULL_XML"))
   {
      pInfo->ivLog->Print("      ERROR: Unable to retrieve XML stored at '@SIMPULL_XML'\n");
      return(IVR_FATAL);
   }
   status = pxmlReq->LoadFromBuffer(pInfo->variables->Get("@SIMPULL_XML"));
   if (status == IVR_SUCCESS)
   {
      // --- Add it to the linked list of open XML documents
      pInfo->xmlList->AddLast((Link*)pxmlReq);
   }

   return(status);
}
// StepSimpullXMLReq::Execute


/* -------------------------------------------------------------------------
   Name:        StepSimpullXMLReq::ComposeParams                   (Private)

   Purpose:     Parse the parameters buffer, removing CR/LF and adding
                special characters required by Simpull XML API.

   Modified:    May 15/02 - TF: Created.

   Notes:       -  Returns char pointer to new string created, or
                   or NULL if an error occurs.
                -  Calling program is responsible for destruction
                   of the string.
   ------------------------------------------------------------------------- */
char* StepSimpullXMLReq::ComposeParams(const char* pszText, PIVRINFO pInfo)
{
   string sComposed;          // composed parameter string
   char*  pszRet = NULL;      // return value, converted from 'string' to 'c_str'
   int    srcpos = 0;         // current position in source string
   string::size_type bufsize;

   if (pszText == NULL || pszText[0] == 0)
   {
      // If no text supplied, return an empty string
      pszRet = new char[1];
      pszRet[0] = '\0';
      return pszRet;
   }

   sComposed = "";

   while (pszText[srcpos] != '\0')
   {
      // Skip white space until we find a parameter
      while (isspace(pszText[srcpos]) && pszText[srcpos] != '\0')
      {
         srcpos++;
      }

      // Check for end of the source string
      if (pszText[srcpos] != '\0')
      {
         // Add the parameter to the destination string
         sComposed += '&';
         while (!isspace(pszText[srcpos]) && pszText[srcpos] != '=' && pszText[srcpos] != '\0')
         {
            sComposed += pszText[srcpos];
            srcpos++;
         } // endwhile

         // Skip white space until we find the equal sign
         while (isspace(pszText[srcpos]) && pszText[srcpos] != '\0')
         {
            srcpos++;
         } // endwhile

         if (pszText[srcpos] == '=')
         {
            sComposed += '=';
            srcpos++;
         } else {
            pInfo->ivLog->Print ("      Illegal syntax - parameters must be followed by '='.\n");
            return(NULL);
         } // endif

         // Copy all characters, ignoring spaces until we get to the end of the line
         while (pszText[srcpos] != '\n' && pszText[srcpos] != '\0')
         {
            if (!isspace(pszText[srcpos]))
            {
               sComposed += pszText[srcpos];
            }
            srcpos++;
         } // endwhile
      } // endif
   } // endwhile

   bufsize = 1;
   bufsize = bufsize + sComposed.length();
   if (bufsize > 0)
   {
      pszRet = new char[bufsize];
      if(pszRet == NULL)
      {
         pInfo->ivLog->Print("      WARNING: Out of memory");
      }
      else
      {
         strcpy(pszRet, sComposed.c_str());
      }
   }
   return pszRet;
}
/*StepSimpullXMLReq::ComposeParams*/


/* -------------------------------------------------------------------------
   Name:        StepSimpullXMLReq::BuildHash                       (Private)

   Purpose:     Build the appropriate MD5 hash string as required by
                Simpull's specifications.

   Modified:    Jul 04/02 - TF: Created.

   Notes:       -  Simpull's specifications are:
                   my $hash = MD5->hexhash($secret .MD5->hexhash(join ':', $secret, $uid, $pid, $t));

                -  Returns char pointer to new string created, or
                   or NULL if an error occurs.

                -  Calling program is responsible for destruction
                   of the string.
   ------------------------------------------------------------------------- */
char* StepSimpullXMLReq::BuildHash (const char* pszSecret,
                                    const char* pszUID,
                                    const char* pszPID,
                                    const char* pszTime)
{
   char*       pszRet = NULL;     // return value
   string      sComposed;         // use STL string to compose values for hashing
   char*       pszHash;           // hash value returned from MD5Hash

   sComposed = pszSecret;
   sComposed += ":";
   sComposed += pszUID;
   sComposed += ":";
   sComposed += pszPID;
   sComposed += ":";
   sComposed += pszTime;

   pszHash = MD5Hash(sComposed.c_str(), false); // Hash in lower case
   if (pszHash != NULL)
   {
      sComposed = pszSecret;
      sComposed += pszHash;
      pszRet = MD5Hash(sComposed.c_str(), false); // Hash in lower case
   }

   delete [] pszHash;
   return pszRet;
}
/*StepSimpullXMLReq::BuildHash*/
