#include "step_vocrecgetstring.h"
#include "..\BoardDLLs\brd_generic.h"
#include "..\Voice\Voice.h"

StepVocRecGetString::StepVocRecGetString(STEPREC& stepRecord) : Step(stepRecord)
{
   grammar = stepRecord.s6.arr;
   grammarText = stepRecord.s8.arr;
   noVoiceTimeout = stepRecord.s3.arr;
   thresholdConfidence = stepRecord.s4.arr;
   confidenceVar = stepRecord.s2.arr;
   resultVar = stepRecord.s1.arr;
   handleVar = stepRecord.s7.arr;
   retainAudioOnThreshold = (bool) stepRecord.u1;
}

StepVocRecGetString::~StepVocRecGetString(void)
{
}

const char* StepVocRecGetString::GetStepName(void) const
{
	return("SR Get String");
}

IV_STATUS StepVocRecGetString::Execute (PIVRINFO pInfo)
{
	IV_STATUS status;
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   int iNoVoiceTimeout;
   unsigned int endOfSpeechTimeout = pInfo->ivBoard->GetSREndOfSpeechTimeout();
   char concept[1024];
   unsigned int maxlength = 1023;
   unsigned int confidence;
   unsigned int iThresholdConfidence;
   string parsedGrammar;
   string parsedGrammarText;
   Voice::SR_GRAMMAR grammarType = Voice::FILE; // assume FILE grammar unless we find out otherwise

	// String large enough for 0 to 100 %
   char confidenceString[4];

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

	if(PortRequired(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FATAL);
   }

   if(pInfo->speechRecognition == NULL)
    {
      pInfo->ivLog->Print("      ERROR: No speech recognition module loaded\n");
      pInfo->ivLog->Print("         Make sure to use SR Acquire before SR steps\n");
      return(IVR_FATAL);
   }

   // Get no voice timeout
   if((value = pInfo->variables->ParseParameter("No voice TIMEOUT", noVoiceTimeout.c_str(), "@TIMEOUT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iNoVoiceTimeout = atoi(value);
      delete[] value;

      if(iNoVoiceTimeout <= 0)
      {
		   pInfo->ivLog->Print("      ERROR: No voice TIMEOUT '%s' must be > 0\n", noVoiceTimeout.c_str());
         bParmsValid = false;
      }
   }

   if(!thresholdConfidence.empty())
   {
      if((value = pInfo->variables->ParseParameter("Threshold confidence", thresholdConfidence.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         iThresholdConfidence = atoi(value);
         delete[] value;
      }
   }
   else
   {
      pInfo->ivLog->Print("      No threshold confidence\n");
   }

   if((value = pInfo->variables->ParseParameter("Grammar", grammar.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedGrammar = value;
      delete[] value;
   }

   if((value = pInfo->variables->ParseParameter("Grammar Text", grammarText.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedGrammarText = value;
      delete[] value;
   }

   // Evaluate grammars
   // -----------------
   if(parsedGrammar.empty() && parsedGrammarText.empty())
   {
      pInfo->ivLog->Print("      ERROR: Either a Grammar or Grammar Text must be specified\n");
      bParmsValid = false;
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   // Blank out variable to receive digits.
   if(!pInfo->variables->Set(resultVar.c_str(), ""))
   {
   	pInfo->ivLog->Print("      ERROR: Unable to clear Response variable '%s'\n", resultVar.c_str());
      return(IVR_FATAL);
   }

   // Capture SR starting information for SpeechTrainer
   pInfo->speechRecognition->SRCaptureStartStats(pInfo->pszAppFile,     // SR step TAP
                                              GetCurrentStep(),         // SR step number
                                              GetStepName(),            // SR step name
                                              iNoVoiceTimeout,          // no voice t/o (int)
                                              endOfSpeechTimeout,       // end speech t/o (unsigned int)
                                              iThresholdConfidence,     // threshold (unsigned int)
                                              retainAudioOnThreshold,   // retainAudio (bool)
                                              grammarType,              // grammar type either FILE or TEXT
                                              parsedGrammar.c_str(),    // base grammar
                                              parsedGrammarText.c_str() // additional grammar
                                            );

   pInfo->speechRecognition->SRSetAudioFormat(pInfo->ivBoard->GetSRAudioFormat());

   if(!parsedGrammar.empty())
   {
      if((status = pInfo->speechRecognition->SRLoadGrammar(Voice::FILE, parsedGrammar.c_str())) != IVR_SUCCESS)
      {
         pInfo->ivLog->Print("      ERROR: Failed to load speech recognition grammar '%s'\n", parsedGrammar.c_str());
         return(status);
      }
      else if(!parsedGrammarText.empty()
         && ((status = pInfo->speechRecognition->SRAddGrammar(Voice::TEXT, parsedGrammarText.c_str())) != IVR_SUCCESS))
      {
         pInfo->ivLog->Print("      ERROR: Failed to load speech recognition grammar '%s'\n", parsedGrammarText.c_str());
         return(status);
      }
   }
   else if((status = pInfo->speechRecognition->SRLoadGrammar(Voice::TEXT, parsedGrammarText.c_str())) != IVR_SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to load speech recognition grammar '%s'\n", parsedGrammarText.c_str());
      return(status);
   }

   // Recognise Audio
  	status = pInfo->ivBoard->AudioRecognise(iNoVoiceTimeout, endOfSpeechTimeout);

   if(status != IVR_SUCCESS)
   {
      pInfo->speechRecognition->SREnd();
      pInfo->speechRecognition->SRClearAudio();
   }

   // No voice detected
	if(status == IVR_NO_VOICE)
   {
      char ttDig;

      if((status = pInfo->ivBoard->QueryDigit(ttDig)) != IVR_SUCCESS)
      {
         return(IVR_FATAL);
      }
      // IVR_TIMEOUT for no voice
      else if(ttDig == '\0')
      {
         // SR Result = NO VOICE
         // Capture SR result for SpeechTrainer
         pInfo->speechRecognition->SRCaptureResult('\0',         // no digit
                                               "",           // no concept text
                                               0,            // confidence undefined
                                               IVR_TIMEOUT); // result

         if(!pInfo->variables->IsEmpty(handleVar.c_str())
            && !pInfo->variables->Set(handleVar.c_str(), pInfo->speechRecognition->SRGetRecognitionHandle()))
         {
            pInfo->ivLog->Print("      ERROR: Failed to set handle variable '%s'\n", handleVar.c_str());
            return(IVR_FATAL);
         }

         return(IVR_TIMEOUT);
      }

      if(!pInfo->variables->IsEmpty(confidenceVar.c_str()))
      {
   	   // Set confidence variable to 0
      	if(!pInfo->variables->Set(confidenceVar.c_str(), "0"))
   	   {
   		   pInfo->ivLog->Print("      ERROR: Failed to set Confidence variable '%s' to '0'\n", confidenceVar.c_str());
	   		return(IVR_FATAL);
   	   }
      }

      // IVR_THRESHOLD for digit
      // SR Result = TT TERMINATION
      // Capture SR result for SpeechTrainer
      pInfo->speechRecognition->SRCaptureResult(ttDig,          // digit
                                            "",             // no concept text
                                            0,              // confidence 0%
                                            IVR_THRESHOLD); // result

      if(!pInfo->variables->IsEmpty(handleVar.c_str())
         && !pInfo->variables->Set(handleVar.c_str(), pInfo->speechRecognition->SRGetRecognitionHandle()))
      {
         pInfo->ivLog->Print("      ERROR: Failed to set handle variable '%s'\n", handleVar.c_str());
         return(IVR_FATAL);
      }

      return(IVR_THRESHOLD);
   }
   // We were unsuccessful with the audio recognition
   else if(status != IVR_SUCCESS)
   {
   	return(status);
   }

   // Get result
   if(pInfo->speechRecognition->SRGetResult(concept, maxlength, confidence) != IVR_SUCCESS)
   {
   	pInfo->ivLog->Print("      ERROR: Failed to get speech recognition result\n");

      // SR Result = SR FAILURE
      // Capture SR result for SpeechTrainer
      pInfo->speechRecognition->SRCaptureResult('\0',           // no digit
                                            "",             // no concept text
                                            0,              // confidence undefined
                                            IVR_FAIL);       // result

      if(!pInfo->variables->IsEmpty(handleVar.c_str())
         && !pInfo->variables->Set(handleVar.c_str(), pInfo->speechRecognition->SRGetRecognitionHandle()))
      {
         pInfo->ivLog->Print("      ERROR: Failed to set handle variable '%s'\n", handleVar.c_str());
         return(IVR_FATAL);
      }

      pInfo->speechRecognition->SRClearAudio();

      return(IVR_FAIL);
   }

   // Set result variable
   if(!pInfo->variables->Set(resultVar.c_str(), concept))
   {
   	pInfo->ivLog->Print("      ERROR: Failed to set Response variable '%s' with '%s'\n", resultVar.c_str(), concept);
      pInfo->speechRecognition->SRClearAudio();
		return(IVR_FATAL);
   }

	if(!pInfo->variables->IsEmpty(confidenceVar.c_str()))
   {
	   // Write confidence to confidence string
   	sprintf(confidenceString, "%d", confidence);

	   // Set confidence variable with confidence
   	if(!pInfo->variables->Set(confidenceVar.c_str(), confidenceString))
   	{
   		pInfo->ivLog->Print("      ERROR: Failed to set Confidence variable '%s' with '%s'\n", confidenceVar.c_str(), confidenceString);
         pInfo->speechRecognition->SRClearAudio();
			return(IVR_FATAL);
   	}
   }

   // Check threshold confidence
	if(!pInfo->variables->IsEmpty(thresholdConfidence.c_str()))
   {
      // Recognition confidence is less than the minimum or empty string
      if((confidence < iThresholdConfidence) || pInfo->variables->IsEmpty(concept))
      {
         // SR Result = THRESHOLD FAILURE
         // Capture SR result for SpeechTrainer
         pInfo->speechRecognition->SRCaptureResult('\0',           // no digit
                                               concept,        // concept text
                                               confidence,     // confidence
                                               IVR_THRESHOLD); // result

         if(!pInfo->variables->IsEmpty(handleVar.c_str())
            && !pInfo->variables->Set(handleVar.c_str(), pInfo->speechRecognition->SRGetRecognitionHandle()))
         {
            pInfo->ivLog->Print("      ERROR: Failed to set handle variable '%s'\n", handleVar.c_str());
            return(IVR_FATAL);
         }

         if(!retainAudioOnThreshold)
         {
            pInfo->speechRecognition->SRClearAudio();
         }

   	   return(IVR_THRESHOLD);
      }
   }

   // SR Result = THRESHOLD PASSED
   // Capture SR result for SpeechTrainer
   pInfo->speechRecognition->SRCaptureResult('\0',           // no digit
                                         concept,        // concept text
                                         confidence,     // confidence
                                         IVR_SUCCESS);    // result

   if(!pInfo->variables->IsEmpty(handleVar.c_str())
      && !pInfo->variables->Set(handleVar.c_str(), pInfo->speechRecognition->SRGetRecognitionHandle()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set handle variable '%s'\n", handleVar.c_str());
      return(IVR_FATAL);
   }

   pInfo->speechRecognition->SRClearAudio();

   return(IVR_SUCCESS);
}
/*StepVocRecGetString::Execute*/
