/* -------------------------------------------------------------------------
   File:        TAP.CPP
   ------------------------------------------------------------------------- */
#include <stdlib.h>             // Needed for "atoi".
#include <dir.h>
#include <io.h>
#include "tap.h"
#include "iv_dbf2.h"
#include "..\escutil\iv_str.h"
#include "step_appcall.h"
#include "step_procstart.h"
#include "eng_main.h"
#include "..\BoardDLLs\brd_generic.h"
#include "..\escutil\esc_except.h"
#include "..\TFSimulator\TFSimulator_communication.h"
#include "..\Voice\Voice.h"
#include "VCE_List.h"
#include "TFEngine_communication.h"
#include "..\TFServer\TCPClientManager.h"
#include "step_vocrecinit.h"

// Numeric defines for application components.
#define  FAXD        "1"
#define  VOIC        "2"
#define  PROC        "3"
#define  LIBR        "4"
#define  NOTE        "5"
#define  STRT        "6"   // Start icon on application screen
#define  APPL        "7"   // Application component used to be '0'

/* *************************************************************************
   *********************** ArgLink Functions *******************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        ArgLink::ArgLink                               (Constructor)
   ------------------------------------------------------------------------- */
ArgLink::ArgLink() : Link()
{
   bUpdateOnReturn = false;
   pszArgNm = 0;
   pszParmNm = 0;
}
/*ArgLink::ArgLink*/

/* -------------------------------------------------------------------------
   Name:        ArgLink::~ArgLink                               (Destructor)
   ------------------------------------------------------------------------- */
ArgLink::~ArgLink()
{
   delete [] pszArgNm;
   delete [] pszParmNm;
}
/*ArgLink::~ArgLink*/

/* -------------------------------------------------------------------------
   Name:        ArgLink::SetArgName

   Purpose:     Copies the portion of 'value' that makes up the parameter
                name into 'pszArgNm'.
   ------------------------------------------------------------------------- */
void ArgLink::SetArgName (const char *value, int len)
{
   SetString (&pszArgNm, value, len);
}
/*ArgLink::SetArgName*/

/* -------------------------------------------------------------------------
   Name:        ArgLink::SetParmName

   Purpose:     Copies the portion of 'value' that makes up the parameter
                name into 'pszParmNm'.
   ------------------------------------------------------------------------- */
void ArgLink::SetParmName (const char *value, int len)
{
   SetString (&pszParmNm, value, len);
}
/*ArgLink::SetParmName*/

/* -------------------------------------------------------------------------
   Name:        ArgLink::SetString

   Purpose:     Copies up to 'len' characters from 'src' into 'dest',
                truncating earlier if a null is encountered in 'src'. 'dest'
                will be null-terminated.
   ------------------------------------------------------------------------- */
void ArgLink::SetString (char **dest, const char *src, int len)
{
   int  iSize;

   if (!dest)
      return;

   if (*dest) {
      delete [] *dest;
      *dest = 0;
   }

   if (src && *src) {
      iSize = strlen (src);
      if (len >= 0 && len < iSize)
         iSize = len;
      *dest = new char [iSize + 1];
      strncpy (*dest, src, iSize);
      (*dest)[iSize] = 0;
   }
}
/*ArgLink::SetString*/

/* *************************************************************************
   ********************** QueueLink Functions ******************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        QueueLink::QueueLink                           (Constructor)

   Purpose:     This class stores information about the current location in
                the current TAM before loading in the next TAM.  It is used
                to restore the current TAM when the next TAM exits.
   ------------------------------------------------------------------------- */
QueueLink::QueueLink() : Link()
{
   pTAM = NULL;               // Pointer to current TAM.
   pStep = NULL;              // Pointer to current step in current TAM.
   pArgList = NULL;           // Link list of parameters used by current TAM.
}
/*QueueLink::QueueLink*/

/* -------------------------------------------------------------------------
   Name:        QueueLink::~QueueLink                           (Destructor)

   Notes:       - The destructor does *not* free memory allocated to the
                  variables stored here.  The pointers will be restored in
                  TAP::PopProcedure and freed in ???.
   ------------------------------------------------------------------------- */
QueueLink::~QueueLink()
{
   delete pArgList;
}
/*QueueLink::~QueueLink*/

/* *************************************************************************
   ************************** TAP Functions ********************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        TAP::TAP                                       (Constructor)

   Purpose:     Initializes the application by loading in the voice scripts,
                loading in all TAMs attached to the TAP file, and identifies
                which TAM is the starting TAM.

   Notes:       - May want to modify 'DefaultExt' to not re-allocate memory
                  but instead include an extra 4 bytes in the size of the
                  application name.
   ------------------------------------------------------------------------- */
TAP::TAP (PIVRINFO pInfo)
{
   ivLog       = pInfo->ivLog;
   TAMs        = 0;
   currentStep = 0;
   currentTAM  = 0;
   qList       = 0;
   bReady      = false;
   startParametersList = new Head();

   memset (&appName, 0, sizeof (MYCHAR));

   if (!pInfo->pszAppFile || !*pInfo->pszAppFile) {
      ivLog->Print ("ERROR: No application name specified.\n");
      return;
   }

   stradd (appName.arr, pInfo->pszAppFile, appName.len);

   // Append default file extension if none supplied.
   DefaultExt (&appName, ".TAP");

   // Create @SYSDRIVE and @SYSDIR variables.
   SetSysVariables(pInfo);

   ivLog->Print ("TeleFlow Application: %-36s \n", appName.arr);
   if (TxtApplHeader (ivLog, appName.arr) != IVR_SUCCESS)
      return;

   if (FirstProcName() == IVR_SUCCESS)
   {
      if (LoadVoiceScripts (pInfo) == IVR_SUCCESS)
      {
         TAMs = new Head();
         qList = new Head();
         if (LoadTAMs (pInfo) == IVR_SUCCESS)
         {
            bReady = true;
         }
      }
   }
}
/*TAP::TAP*/

/* -------------------------------------------------------------------------
   Name:        TAP::~TAP                                       (Destructor)

   Purpose:     Frees all memory used the the application.  All memory used
                by the TAMs (and all memory used by steps attached to TAMs)
                is released.
   ------------------------------------------------------------------------- */
TAP::~TAP()
{
   delete [] appName.arr;
   delete TAMs;
   delete qList;
   delete startParametersList;
}
/*TAP::~TAP*/

bool TAP::GetReady(void)
{
   return(bReady);
}

/* -------------------------------------------------------------------------
   Name:        TAP::SetSysVariables
   ------------------------------------------------------------------------- */
void TAP::SetSysVariables(PIVRINFO pInfo)
{
   char  szDrive[MAXDRIVE];
   char  szDir[MAXDIR];

   fnsplit(appName.arr, szDrive, szDir, 0, 0);

   // Remove '\'s from end of szDir
   if(szDrive[strlen(szDrive) - 1] == '\\')
   {
   	szDrive[strlen(szDrive) - 1] = '\0';
   }
   if(szDir[strlen(szDir) - 1] == '\\')
   {
	   szDir[strlen(szDir) - 1] = '\0';
   }
/*
   int   iDirLen;
   iDirLen = strlen (szDir);
   if (iDirLen > 1)
      szDir[iDirLen - 1] = 0;
*/
   pInfo->variables->SetSystem("@SYSDRIVE", szDrive);
   pInfo->variables->SetSystem("@SYSDIR", szDir);
}
/*TAP::SetSysVariables*/

IV_STATUS TAP::LoadParameters(PIVRINFO pInfo)
{
   FILE* TAPFile;
   IV_STATUS status;
   MYCHAR value;

   memset(&value, 0, sizeof(MYCHAR));
   startParametersList->Clear();

   // Open application file.
   if((status = IvOpenFile(ivLog, &TAPFile, appName.arr)) != IVR_SUCCESS)
   {
      return(status);
   }

   TxtSkipSection(TAPFile);    // Skip past intro / notes.
   TxtSkipSection(TAPFile);    // Skip past header.

   // Search for TAP start information
   while(!feof(TAPFile) && ((TxtReadString(TAPFile, &value) < 0) || (strcmp(value.arr, STRT) != 0)))
   {
      TxtReadString(TAPFile, &value);
   }

   if(!feof(TAPFile))
   {
      char* subFieldStart = NULL;
      char* subFieldEnd = NULL;

      // Skip to parameters field
      for(unsigned int counter = 0; counter < 17; counter++)
      {
         TxtReadString(TAPFile, &value);
      }

      TxtReadString(TAPFile, &value);

      // Initialize pointers used to parse to parameters fields from 'value'
      subFieldStart = value.arr;
      subFieldEnd   = value.arr;

      for(unsigned int index = 1;
         (status == IVR_SUCCESS) && (subFieldStart[0] != '\0') && (subFieldEnd != NULL);
         index++)
      {
         char parameterValue[8];
         string parameterName;
         string parameterType;

         sprintf(parameterValue, "@P%d", index);

         subFieldEnd = strchr(subFieldStart, END_SUBFIELD);

         if(subFieldEnd != NULL)
         {
            parameterName.assign(subFieldStart, 0, subFieldEnd - subFieldStart);
            subFieldStart = subFieldEnd + 1;

            if(subFieldStart[0] != '\0')
            {
               subFieldEnd = strchr(subFieldStart, END_SUBFIELD);

               if(subFieldEnd != NULL)
               {
                  parameterType.assign(subFieldStart, 0, subFieldEnd - subFieldStart);
                  subFieldStart = subFieldEnd + 1;

                  if(!parameterName.empty())
                  {
                     if(parameterType.empty())
                     {
                        ivLog->Print("ERROR: TAP parameter %d '%s' has no value/reference designation\n", index, parameterName.c_str());
                        IvCloseFile(ivLog, &TAPFile, appName.arr);
                        status = IVR_FATAL;
                     }
                     else
                     {
                        bool bReference = (atoi(parameterType.c_str()) != 0);
                        ArgLink* parameterLink = new ArgLink();

                        parameterLink->SetArgName(parameterName.c_str());
                        parameterLink->SetUpdateOnReturn(bReference);
                        startParametersList->AddLast((Link*) parameterLink);
                        // Show the mapping between @P# and named parameters
                        ivLog->Print("      %s (pass by %s) = %s\n", parameterName.c_str(), (bReference?"reference":"value"), parameterValue);
                        if(!pInfo->variables->Set(parameterName.c_str(), pInfo->variables->Get(parameterValue)))
                        {
                           ivLog->Print("ERROR: Failed to set TAP parameter %d '%s'\n", index, parameterName.c_str());
                           IvCloseFile(ivLog, &TAPFile, appName.arr);
                           status = IVR_FATAL;
                        }
                        else
                        {
                           subFieldStart = subFieldEnd + 1;
                           subFieldEnd = strchr(subFieldStart, END_SUBSECT);
                           if(subFieldEnd != NULL)
                           {
                              subFieldStart = subFieldEnd + 1;
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }

   delete[] value.arr;

   if(IvCloseFile(ivLog, &TAPFile, appName.arr) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   return(status);
}

IV_STATUS TAP::SetReferenceParameters(PIVRINFO pInfo, PIVRINFO parentInfo, const char* parameters)
{
   const char* parentParameterNameStart = parameters != NULL ? parameters : "";
   const char* parentParameterNameEnd = strchr(parentParameterNameStart, END_SUBFIELD);
   ArgLink* parameterLink = (ArgLink*) startParametersList->First();

   for(unsigned int index = 1; parameterLink != NULL; index++)
   {
      if(parameterLink->GetUpdateOnReturn())
      {
         string parentParameter;

         if(parentParameterNameEnd == NULL)
         {
            pInfo->ivLog->Print("      ERROR: Missing reference parameter @P%d\n", index);
            return(IVR_FATAL);
         }

         parentParameter.assign(parentParameterNameStart, 0, parentParameterNameEnd - parentParameterNameStart);

         if(!parentInfo->variables->Set(parentParameter.c_str(), pInfo->variables->Get(parameterLink->GetArgName())))
         {
            pInfo->ivLog->Print("      ERROR: Failed to set reference parameter @P%d '%s'\n", index, parentParameter.c_str());
            return(IVR_FATAL);
         }
      }

      if(parentParameterNameEnd != NULL)
      {
         parentParameterNameStart = parentParameterNameEnd + 2;
         parentParameterNameEnd = strchr(parentParameterNameStart, END_SUBFIELD);
      }

      parameterLink = (ArgLink*) parameterLink->Suc();
   }

   return(IVR_SUCCESS);
}

#if 0
/* -------------------------------------------------------------------------
   Name:        TAP::StartMonReporter
   ------------------------------------------------------------------------- */
IV_STATUS TAP::StartMonReporter (PIVRINFO pInfo)
{
   IV_STATUS  status;

   // Initialize structure used to send messages to monitor application.
   if ((status = IPCInitialize (&info)) != IVR_SUCCESS)
      return status;

   // Initialize structure to synchronize communication with monitor thread.
   mySem = new ipcSemSemaphore (CLIENT_SEM_NAME, semcreate);
   if (mySem->Query())
      mySem->Release();

   // Initialize information to be used by the monitor thread.
   memset (&cs, 0, sizeof (CLIENT_STRUCT));
   cs.appID = pInfo->ivAppId;
   sprintf (cs.semName, "%s%d", CLIENT_SEM_NAME, cs.appID);

   hThread = CreateThread (
                NULL,               // Security flags.
                1024,               // Stack size.
                MonReporter,        // Thread function.
                (LPVOID)&cs,        // Parameter to function.
                0,                  // Creation flags.
                lpIDThread);        // Thread identifier.

   if (!hThread) {
      printf ("Failed to create thread.\n");
      IPCShutdown (&info);
      delete mySem;
      return;
   }

   return IVR_SUCCESS;
}
/*TAP::StartMonReporter*/

/* -------------------------------------------------------------------------
   Name:        TAP::GetMonCommand
   ------------------------------------------------------------------------- */
void TAP::GetMonCommand (IV_COMMAND &cmd, CMD_LEVEL &lvl)


{
   // Check to see if monitor updated status variables->
   mySem->Request (false, 5);
   if (mySem->Query()) {
      cmd = cs.cmd;
      lvl = cs.lvl;
   }
   mySem->Release();
}
/*TAP::GetMonCommand*/

/* -------------------------------------------------------------------------
   Name:        TAP::SetMonStep
   ------------------------------------------------------------------------- */
void TAP::SetMonStep ()
{
   cs.cMsg.appID =
}
/*TAP::SetMonStep*/
#endif

/* -------------------------------------------------------------------------
   Name:        TAP::LoadVoiceScripts

   Purpose:     Loads in the voice scripts.

                A voice script is a code used by the application designer to
                reference an audio file and possibly a specific message in
                the voice file.  If the application designer uses a voice
                script code, the code must be used to look up a file name to
                play.
   ------------------------------------------------------------------------- */
IV_STATUS TAP::LoadVoiceScripts (PIVRINFO pInfo)
{
   IV_STATUS  status;
   bool       bContinue = true;
   int        pos;
   int        counter;
   FILE      *fApp;
   DBFFile    dbFile;
   VoiceLink  vLink;
   string     scriptCodeName;
   MYCHAR     value;
   MYCHAR     voiceFile;
   MYCHAR     voiceSearchPath;

   memset (&value, 0, sizeof (MYCHAR));
   memset (&voiceFile, 0, sizeof (MYCHAR));
   memset (&voiceSearchPath, 0, sizeof (MYCHAR));
   pInfo->pszVoiceSearchPath = NULL;

   if ((status = IvOpenFile (ivLog, &fApp, appName.arr)) != IVR_SUCCESS)
      return status;

   TxtSkipSection (fApp);      // Skip past intro/notes
   TxtSkipSection (fApp);      // Skip past header

   while (!feof (fApp) && bContinue) {
      bContinue = false;
      pos = TxtReadString (fApp, &value);
      if (pos >= 0) {
         if (!strcmp (value.arr, VOIC))
         {
            status = IVR_FAIL;
            for (counter = 2; counter <= 13; ++counter)
            {
               TxtReadString (fApp, &value);
            }
            pos = TxtReadString (fApp, &voiceFile);
            if (pos >= 0)
            {
               // found voice script DB file name
               status = IVR_SUCCESS;

               // also read the optional voice file search path
               pos = TxtReadString (fApp, &voiceSearchPath);
               if (pos > 0)
               {
                  pInfo->pszVoiceSearchPath = new char[pos + 1];
                  strcpy(pInfo->pszVoiceSearchPath, voiceSearchPath.arr);
                  delete [] voiceSearchPath.arr;
               }
            }
         }
         else
         {
            if ((status = TxtNextRecord (fApp)) == IVR_SUCCESS)
               bContinue = true;
            else
               status = IVR_SUCCESS;
         }
      }
   }

   if (IvCloseFile (ivLog, &fApp, appName.arr) != IVR_SUCCESS)
   {
      delete [] voiceFile.arr;
      delete [] value.arr;
      return IVR_FAIL;
   }

   if(status == IVR_SUCCESS)
   {
      char* pathName = pInfo->variables->FilePath(voiceFile.arr);
      delete[] voiceFile.arr;
      voiceFile.arr = pathName;
      voiceFile.len = strlen(voiceFile.arr);
      pInfo->ivLog->Print("TeleFlow Voice Database: %s\n", voiceFile.arr);
      if (!dbFile.Open(voiceFile.arr))
      {
         ivLog->Print ("ERROR: Failed to open voice file '%s'\n",
                       voiceFile.arr);
         ivLog->Print ("Possible reasons for failure:\n"
                       "1) File is not a dBASE III file (Don't modify file"
                         " with third party tool!).\n"
                       "2) File is read-only"
                         " (use ATTRIB -R to remove read-only flag).\n");
         delete[] voiceFile.arr;
         delete[] value.arr;
         return(IVR_FAIL);
      }
      if(pInfo->pszVoiceSearchPath != NULL)
      {
         pInfo->ivLog->Print("   Voice file search path: %s\n", pInfo->pszVoiceSearchPath);
      }
      else
      {
         pInfo->ivLog->Print("   Voice file search path: None\n");
      }

      for (pos = 1; dbFile.Fetch (pos); ++pos)
      {
         if(!dbFile.IsRecordDeleted())
         {
            strclear (value.arr);
            char* fieldValue = new char[dbFile.GetFieldValueLength("NAME") + 1];
            stradd(value.arr, dbFile.GetFieldValue(fieldValue, "NAME"), value.len);
            delete[] fieldValue;
            strtrim_right (value.arr);
            scriptCodeName = value.arr;

            // TFSimulator uses the WAVFILE field.  It can only handle .WAV files.
            strclear (value.arr);
            fieldValue = new char[dbFile.GetFieldValueLength("WAVFILE") + 1];
            stradd(value.arr, dbFile.GetFieldValue(fieldValue, "WAVFILE"), value.len);
            delete[] fieldValue;
            strtrim_right (value.arr);

            value.len = strlen (value.arr);
            vLink.SimulatorFileName = value.arr;

            // Telephony Engine uses the VOXFILE field.  Any audio format.
            strclear (value.arr);
            fieldValue = new char[dbFile.GetFieldValueLength("VOXFILE") + 1];
            stradd(value.arr, dbFile.GetFieldValue(fieldValue, "VOXFILE"), value.len);
            delete[] fieldValue;
            strtrim_right (value.arr);

            value.len = strlen (value.arr);
            vLink.TelephonyFileName = value.arr;

            strclear (value.arr);
            fieldValue = new char[dbFile.GetFieldValueLength("VOXCONV") + 1];
            stradd(value.arr, dbFile.GetFieldValue(fieldValue, "VOXCONV"), value.len);
            delete[] fieldValue;
            vLink.Format = (SOUND_FORMAT) atoi(value.arr);

            pInfo->voiceList->Add(scriptCodeName.c_str(), vLink);
         }
      }
   }

   delete [] voiceFile.arr;
   delete [] value.arr;
   return status;
}
/*TAP::LoadVoiceScripts*/

/* -------------------------------------------------------------------------
   Name:        TAP::LoadTAMs

   Purpose:     Load procedures and steps into linked list.
   ------------------------------------------------------------------------- */
IV_STATUS TAP::LoadTAMs (PIVRINFO pInfo)
{
   FILE      *fApp;
   TAM       *tam;
   bool       bContinue = true;
   IV_STATUS  status;
   int        pos;
   int        counter;
   int        iTAMID;
   MYCHAR     value;
   MYCHAR     tamFile;
   MYCHAR     tamName;
   bool       logDisabled;

   memset (&value,   0, sizeof (MYCHAR));
   memset (&tamFile, 0, sizeof (MYCHAR));
   memset (&tamName, 0, sizeof (MYCHAR));

   // Open application file.
   if ((status = IvOpenFile (ivLog, &fApp, appName.arr)) != IVR_SUCCESS)
      return status;

   TxtSkipSection (fApp);    // Skip past intro / notes.
   TxtSkipSection (fApp);    // Skip past header.

   while (!feof (fApp) && bContinue)
   {
      bContinue = false;
      pos = TxtReadString (fApp, &value);                   // OP code field.
      if (pos >= 0)
      {
         if(!strcmp (value.arr, PROC))
         {
            status = IVR_FAIL;
            pos = TxtReadString (fApp, &tamName);           // Label field.
            if (pos >= 0)
            {
               TxtReadString (fApp, &value);                // Step no field.
               sscanf (value.arr, "%d", &iTAMID);

               // Skip next 2 fields
               TxtReadString (fApp, &value);
               TxtReadString (fApp, &value);

               // Get log disable flag
               TxtReadString(fApp, &value);
               logDisabled = strcmp(value.arr, "1") == 0;

               for (counter = 7; counter < 14; ++counter) // Skip to 14th field
               {
                  TxtReadString (fApp, &value);
               }

               pos = TxtReadString (fApp, &tamFile);
               if (pos >= 0)
               {
                  char* pathName;
                  DefaultExt (&tamFile, ".TAM");
                  if (iFirstProcID == iTAMID)
                     ivLog->Debug ('p', "TeleFlow Procedure: > %-36s",
                                   tamFile.arr);
                  else
                     ivLog->Debug ('p', "TeleFlow Procedure:   %-36s",
                                   tamFile.arr);
                  pathName = pInfo->variables->FilePath(tamFile.arr);
                  delete[] tamFile.arr;
                  tamFile.arr = pathName;
                  tamFile.len = strlen (tamFile.arr);
                  if (!FileExist (tamFile.arr))
                     ivLog->Print ("Failed to find procedure file '%s'.\n",
                                  tamFile.arr);
                  else
                  {
                     tam = new TAM(ivLog, logDisabled, tamFile.arr, pInfo->iPort);
                     if(tam)
                     {
                        tam->SetTAMID (iTAMID);
                        TAMs->AddLast ((Link*)tam);
                        if(tam->GetState() == TAM::good)
                        {
                           bContinue = true;
                           status = IVR_SUCCESS;
                        }
                        TxtNextRecord(fApp);
                     }
                     else
                     {
                        ivLog->Print ("ERROR: Out of memory!\nFailed attempting to load TAM.\n");
                     }
                  }
               }
            }
         } else {                            // Not a procedure record skip it.
            if ((status = TxtNextRecord (fApp)) == IVR_SUCCESS)
               bContinue = true;
            else
               status = IVR_SUCCESS;           // Ignore errors on skip record.
         }
      }
   }
   delete [] value.arr;
   delete [] tamFile.arr;
   delete [] tamName.arr;

   if (IvCloseFile (ivLog, &fApp, appName.arr) != IVR_SUCCESS)
      return IVR_FAIL;

   return status;
}
/*TAP::LoadTAMs*/

/* -------------------------------------------------------------------------
   Name:        TAP::PushProcedure

   Notes:       - The current step will be AppCall.
                - Need to know info about the ProcStart and AppCall step at
                  the same time.
                * *** I believe I'm using parameter and argument with
                  *** reverse meanings.
   ------------------------------------------------------------------------- */
IV_STATUS TAP::PushProcedure (PIVRINFO pInfo)
{
   Head* argList;
   ArgLink* argLink;
   QueueLink* qLink;
   TAM* nextTAM;
   StepAppCall* appCall;
   StepProcStart* procStart;
   const char* pszListStart;
   const char* pszListEnd;
   char szNum[10];
   char* value;

   // Current step is AppCall.  Need pointer to access it's member functions.
   appCall = (StepAppCall*) currentStep;

   // Locate the ProcStart icon.  Need to access it's member functions.
   nextTAM = FindTAM (currentStep->GetNextTAMName());
   if (nextTAM)
      procStart = (StepProcStart*) nextTAM->GetFirstStep();
   else
      procStart = 0;

   if (!procStart) {
      ivLog->Print ("ERROR: Failed to find the first step of "
                   "flowchart '%s'.\n", currentStep->GetNextTAMName());
      return IVR_FATAL;
   }

   argList = new Head();
   argLink = 0;

   // -------------------------------------------------------------------------
   // Build a link list storing the procedures arguments.
   // -------------------------------------------------------------------------
   pszListStart = procStart->GetArgList();
   while(pszListStart && *pszListStart)
   {
      pszListEnd = strchr (pszListStart, END_SUBFIELD);
      if(pszListEnd)
      {
         argLink = new ArgLink();
         argLink->SetArgName(pszListStart, pszListEnd - pszListStart);
         pszListStart = pszListEnd + 1;
      }
      if(*pszListStart && pszListEnd)
      {
         pszListEnd = strchr (pszListStart, END_SUBFIELD);
         if (pszListEnd && pszListEnd - pszListStart < 10) {
            strncpy (szNum, pszListStart, pszListEnd - pszListStart);
            szNum[1] = 0;
            argLink->SetUpdateOnReturn (atoi (szNum) ? true : false);
         }
         argList->AddLast ((Link*)argLink);
         argLink = 0;
      }
      delete argLink; // Just in case one is created and not added to the list.
      argLink = 0;
      pszListEnd = strchr (pszListStart, END_SUBSECT);
      if (pszListEnd)
         pszListStart = pszListEnd + 1;
      else
         pszListStart = 0;
   }

   argLink = (ArgLink*) argList->First();

   // -------------------------------------------------------------------------
   // Update the link list with the calling procedures parameters.
   // -------------------------------------------------------------------------
   pszListStart = appCall->GetParmList();
   while(argLink && pszListStart && *pszListStart)
   {
      pszListEnd = strchr(pszListStart, END_SUBFIELD);
      if(pszListEnd)
      {
         argLink->SetParmName (pszListStart, pszListEnd - pszListStart);
         pszListStart = pszListEnd + 1;
      }
      argLink = (ArgLink*) argLink->Suc();
      pszListEnd = strchr (pszListStart, END_SUBSECT);
      if(pszListEnd)
      {
         pszListStart = pszListEnd + 1;
      }
      else
      {
         pszListStart = 0;
      }
   }

   // -------------------------------------------------------------------------
   // Save all the information needed to restore the procedure to its current
   // state when it returns from the calling procedure.
   // -------------------------------------------------------------------------
   qLink = new QueueLink();
   qLink->SetTAM (currentTAM);
   qLink->SetStep (currentTAM->GetStep (currentStep->GetNextStep()));
   pInfo->variables->PushLocalVariables();
   qLink->SetArgList(argList);

   // -------------------------------------------------------------------------
   // Build a new private variable link list for the called procedure.
   // -------------------------------------------------------------------------
   argLink = (ArgLink*) argList->First();
   while(argLink)
   {
      value = pInfo->variables->ParseWithPushedVariables(argLink->GetParmName());
      if(value == NULL)
      {
         delete qLink;
         return(IVR_FATAL);
      }

      if(!pInfo->variables->SetLocal(argLink->GetArgName(), value))
      {
         delete[] value;
         delete qLink;
         ivLog->Print ("ERROR: Failed to set procedure variable '%s'\n", argLink->GetParmName());

         return(IVR_FATAL);
      }

      delete[] value;
      argLink = (ArgLink*) argLink->Suc();
   }

   qList->AddFirst(qLink);


   if(pInfo->speechRecognition != NULL)
   {
      pInfo->speechRecognition->SRUpdateTAMInfo(nextTAM->GetName());
   }

   return(IVR_SUCCESS);
}
/*TAP::PushProcedure*/

/* -------------------------------------------------------------------------
   Name:        TAP::PopProcedure

   Purpose:     Restores the previous procedure.  If the procedure specifies
                parameters passed by reference, the arguments passed in are
                updated when the TAM exits.
   ------------------------------------------------------------------------- */
IV_STATUS TAP::PopProcedure(PIVRINFO pInfo)
{
   Head* argList;
   ArgLink* argLink;
   QueueLink* qLink;
   const char* leavingTAM = currentTAM->GetShortName();

   qLink = (QueueLink*)qList->First();
   if(qLink == NULL)
   {
      return(IVR_FAIL);
   }

   ivLog->Disable(false);
   qLink->Out();
   currentTAM  = qLink->GetTAM();
   ivLog->Print("%s<==%s\n", currentTAM->GetShortName(), leavingTAM);
   if(pInfo->speechRecognition != NULL)
   {
      pInfo->speechRecognition->SRUpdateTAMInfo(currentTAM->GetName());
   }
   ivLog->Disable(currentTAM->IsLogDisabled());
   currentStep = qLink->GetStep();
   argList = qLink->GetArgList();           // Head of link list. Never NULL.
   argLink = (ArgLink*) argList->First();
   while(argLink)
   {
      if(argLink->GetUpdateOnReturn())
      {
         // Gets the value assigned to the *parameter*.
         char* value = pInfo->variables->Parse(argLink->GetArgName());
         if(value == NULL)
         {
            // Restore local variables before returning.
            pInfo->variables->PopLocalVariables();
            delete qLink;
            return(IVR_FATAL);
         }

         // Assigns the value to the *argument*.
         if(!pInfo->variables->SetPopVariable(argLink->GetParmName(), value))
         {
            ivLog->Print ("ERROR: Unable to update arguments passed to procedure\n");
            ivLog->Print ("  1) Check to see if parameter is a variable if"
                         " arguments passed by reference\n");
            ivLog->Print ("  2) Check parameter list: Did you supply enough"
                         " arguments?\n");
            // Restore local variables before returning.
            pInfo->variables->PopLocalVariables();
            delete qLink;
            delete[] value;
            return(IVR_FATAL);
         }
         delete[] value;
      }
      argLink = (ArgLink*) argLink->Suc();
   }

   // Restore private var. list and free memory before returning.
   pInfo->variables->PopLocalVariables();
   delete qLink;
   return(IVR_SUCCESS);
}
/*TAP::PopProcedure*/

/* -------------------------------------------------------------------------
   Name:        TAP::FindTAMID

   Purpose:     Finds a TAM using it's id.

   Notes:       - This procedure was needed to fix a bug in the Designer.
                  All icons in the Designer have a step number assigned to
                  them and special code was needed to point to the first TAM
                  in the TAP file.  Using step numbers makes that special
                  code unnecessary.
   ------------------------------------------------------------------------- */
TAM * TAP::FindTAMID (int id)
{
   TAM  *temp;

   if (!TAMs || !id) {
      ivLog->Print ("not found.\n");
      return 0;
   }

   temp = (TAM*)TAMs->First();
   while (temp) {
      if (temp->GetTAMID() == id)
         return temp;
      temp = (TAM*)temp->Suc();
   }
   return 0;
}
/*TAP::FindTAMID*/

/* -------------------------------------------------------------------------
   Name:        TAP::FindTAM

   Purpose:     Finds a TAM using it's name.

   Notes:       - This function is no longer used.
   ------------------------------------------------------------------------- */
TAM * TAP::FindTAM (const char *name)
{
   TAM        *temp;

   if (!TAMs || !name || !*name) {
      ivLog->Print ("not found.\n");
      return 0;
   }

   temp = (TAM*)TAMs->First();
   while (temp) {
      if (!stricmp (temp->GetShortName(), name)) {
         return temp;
      }
      temp = (TAM*)temp->Suc();
   }
   return 0;
}
/*TAP::FindTAM*/

/* -------------------------------------------------------------------------
   Name:        TAP::Execute

   Notes:       - I really hate referencing IvBoard member functions from
                  function, but I don't want to include VAR.CPP when
                  building IV_BOARD.CPP - JR
   ------------------------------------------------------------------------- */
IV_STATUS TAP::Execute (PIVRINFO pInfo)
{
   IV_STATUS   status;
   bool        bContinue = true;
   bool        bOnHookDelta;

   // Find the first TAM module.
   currentTAM = FindTAMID (iFirstProcID);
   if (!currentTAM) {
      ivLog->Print ("ERROR: Starting Point not connected to first flowchart\n");
      return IVR_FAIL;
   }

   // Find the first step in that TAM module.
   currentStep = currentTAM->GetFirstStep();
   if(currentStep == NULL)
   {
      ivLog->Print ("   ERROR: Failed to find first step in TAM\n");
      return IVR_FAIL;
   }
   ivLog->Print("==>%s\n", currentTAM->GetShortName());
   if(currentTAM->IsLogDisabled())
   {
      ivLog->Print("   Logging disabled in this flowchart\n");
   }
   ivLog->Disable(currentTAM->IsLogDisabled());

   // Line always on-hook when application starts.
   bOnHookDelta = true;
   while(bContinue && (currentStep != NULL))
   {
      SetMonData(pInfo);
      SendSimulatorData(pInfo);

      ivLog->Print("   [%d] %s\n", currentStep->GetCurrentStep(), currentStep->GetStepName());

      // This function is used by Similator board to support the 'Pause' button
      // For telephony boards, it will return immediately
      pInfo->ivBoard->WaitWhilePaused();

      status = currentStep->Execute (pInfo);

      if (bOnHookDelta && !pInfo->ivBoard->GetLineOnHook()) {
         // Call in progress, initialize call stats.
         // Phone is taken off hook in the ANSWERPHONE and CALLACCEPT steps.
         bOnHookDelta = false;

      } else if (!bOnHookDelta && pInfo->ivBoard->GetLineOnHook()) {
         // Caller hung up or call terminated, set call stats.
         bOnHookDelta = true;
      }

      switch (status) {
         case IVR_SUCCESS: // -------------------------------------------------
            currentStep = currentTAM->GetStep (currentStep->GetNextStep());
            break;

         case IVR_FAIL: // ----------------------------------------------------
            currentStep = currentTAM->GetStep (currentStep->GetFailureStep());
            ivLog->Print("      FAILURE\n");
            break;

         case IVR_TRUE:
         {
         	currentStep = currentTAM->GetStep(currentStep->GetTrueStep());
            ivLog->Print("      TRUE\n");

            break;
         }

         case IVR_FALSE:
         {
         	currentStep = currentTAM->GetStep(currentStep->GetFalseStep());
            ivLog->Print("      FALSE\n");

            break;
         }

         case IVR_THRESHOLD:
         {
         	currentStep = currentTAM->GetStep(currentStep->GetThresholdStep());
            ivLog->Print("      THRESHOLD\n");

            break;
         }

         case IVR_TIMEOUT: // -------------------------------------------------
            currentStep = currentTAM->GetStep (currentStep->GetTimeoutStep());
            ivLog->Print("      TIMEOUT\n");
            break;

         case IVR_NO_VOICE: // -------------------------------------------------
         {
            currentStep = currentTAM->GetStep(currentStep->GetNoVoiceStep());
            ivLog->Print("      NO VOICE\n");
            break;
         }

         case IVR_SILENCE_TIMEOUT: // -------------------------------------------------
         {
            currentStep = currentTAM->GetStep(currentStep->GetSilenceStep());
            ivLog->Print("      SILENCE\n");
            break;
         }

         case IVR_BUSY:    // -------------------------------------------------
            currentStep = currentTAM->GetStep (currentStep->GetBusyStep());
            ivLog->Print("      BUSY\n");
            break;

         case IVR_HANGUP: // --------------------------------------------------
         {
            char* value = pInfo->variables->Parse("@HANGUP");

            ivLog->Print("      HANGUP\n");
            ivLog->Print("         Attempting to jump to Hangup Global Event label: '%s'\n", value);

            if(pInfo->variables->IsEmpty(value))

            {
               ivLog->Print("            Hangup label undefined!\n");
               bContinue = false;
               currentStep = NULL;     // Ensure a crash is logged - TF Mar.7/02
               status = IVR_FATAL;
            }
            else
            {
               currentStep = currentTAM->GetLabelStep(value);

               // If IVR_HANGUP didn't find a step in current TAM, pop the call
               // stack looking for the label.
               while(bContinue && (currentStep == NULL) && (status != IVR_FATAL))
               {
                  if(PopProcedure(pInfo) == IVR_SUCCESS)
                  {
                     currentStep = currentTAM->GetLabelStep(value);
                  }
                  else
                  {
                     ivLog->Print("            Could not locate Hangup Global Event label: '%s'\n", value);
                     status = IVR_FATAL;
                  }
               }
            }

            delete[] value;
            break;
         }

			case IVR_NOTSUPPORTED:
         {
            ivLog->Print("      This feature is not supported in this software module.\n"
            				 "      Please contact Engenic for more information.\n");
         }	// Fall through to IVR_FATAL

         case IVR_FATAL: // ---------------------------------------------------
         {
            char* value = pInfo->variables->Parse("@FATAL");

            // Flag monitor that we have crashed and allow system to know we
            // crashed even if the FATAL event handler is successful
   			pInfo->appFlags.status = lsCrashed;

            ivLog->Disable(false);
            ivLog->Print("      FATAL\n");

            ivLog->Print("         Attempting to jump to Fatal Error Global Event label: '%s'\n", value);

            if(pInfo->variables->IsEmpty(value))
            {
               ivLog->Print("         Fatal label undefined!\n");
               while(PopProcedure(pInfo) == IVR_SUCCESS)
               {
                  // Discard results, simply un-wind procedure calls so we can exit safely
               }
               bContinue = false;
               currentStep = NULL;
            }
            else
            {
               currentStep = currentTAM->GetLabelStep(value);

               // If IVR_FATAL didn't find a step in current TAM, pop the call
               // stack looking for the label.
               while(bContinue && (currentStep == NULL))
               {
                  if(PopProcedure (pInfo) == IVR_SUCCESS)
                  {
                     currentStep = currentTAM->GetLabelStep (value);
                  }
                  else
                  {
                     ivLog->Print("            Could not locate Fatal Error Global Event label: '%s'\n", value);
                     bContinue = false;
                  }
               }

               pInfo->variables->SetSystem("@FATAL", "");
            }

            delete[] value;
            break;
         }

         case IVR_SHUTDOWN: // ------------------------------------------------
            ivLog->Print("      SHUTDOWN\n");
            while (PopProcedure (pInfo) == IVR_SUCCESS)
               ;  // If shutting down, release allocated memory.
            bContinue = false;
            // Note: do not log a crash on SHUTDOWN events - TF Mar.7/02
            break;

         case IVR_ENDPROC: // -------------------------------------------------
         {
            status = PopProcedure(pInfo);
            if(status != IVR_SUCCESS)
            {
               bContinue = false;
					currentStep = NULL;
               if(status == IVR_FAIL)
               {
                  ivLog->Print("End of application.\n");
                  status = IVR_ENDPROC;
               }
            }
            break;
         }

         case IVR_APPCALL: // -------------------------------------------------
         {
            // Want bad procedure calls to trigger fatal error - TF Mar.5/02
            ivLog->Disable(false);
            status = PushProcedure (pInfo);
            if (status != IVR_SUCCESS)
            {
					currentStep  = 0;
               break;
            }

            // Find the requested TAM.
            if(bContinue)
            {
               ivLog->Print("%s==>", currentTAM->GetShortName());

               currentTAM = FindTAM (currentStep->GetNextTAMName());

               if(currentTAM != NULL)
               {
                  ivLog->Print("%s\n", currentTAM->GetShortName());
                  currentStep = currentTAM->GetFirstStep();
               }
               else
               {
                  ivLog->Print("\n      ERROR: Flowchart '%s' not found.\n", currentStep->GetNextTAMName());
                  ivLog->Print("      HINT: Referenced flowcharts must be included on the main page of your\n"
                               "      application and the flowchart file must exist on disk.\n");
                  currentStep = 0;
               }

               if(currentTAM->IsLogDisabled())
               {
                  ivLog->Print("   Logging disabled in this flowchart\n");
                  ivLog->Disable(true);
               }
            }

            break;
         }

         case IVR_GOTO: // ----------------------------------------------------
         {
            const char *nextLabel;

            nextLabel = currentStep->GetNextLabel();
            if (!nextLabel || !*nextLabel)
            {
               ivLog->Print ("      ERROR: No label defined.\n");
               bContinue = false;
               currentStep = 0;     // Ensure a crash is logged - TF Mar.7/02
            } else
               currentStep = currentTAM->GetLabelStep (nextLabel);

            while (bContinue && (currentStep == NULL))
            {
               if (PopProcedure (pInfo) == IVR_SUCCESS)
               {
                  currentStep = currentTAM->GetLabelStep (nextLabel);
               }
               else
               {
                  bContinue = false;
                  currentStep = 0;     // Ensure a crash is logged - TF Mar.7/02
               }
            }
            break;
         }
         default: // ----------------------------------------------------------
            ivLog->Print ("TAP::Execute: unhandled status (%d)\n",
                          status);
            bContinue = false;
            currentStep = 0;     // Ensure a crash is logged - TF Mar.7/02
      }

      // Failed to find the next step.  Jump to the fatal error handler.
      if((currentStep == NULL) && (status != IVR_ENDPROC))
      {
         // If we haven't already "crashed", then we really are missing
         // the next step.  Otherwise this indicates we did not have a
         // fatal error handler defined. - TF Jul.26/02
         if (pInfo->appFlags.status != lsCrashed)
         {
            char* value = pInfo->variables->Parse("@FATAL");

            // Flag monitor that we have crashed and allow system to know we
            // crashed even if the FATAL event handler is successful
   			pInfo->appFlags.status = lsCrashed;

            ivLog->Print("      ERROR: Missing next step\n");
            ivLog->Print("      FATAL\n");
            ivLog->Print("         Attempting to jump to Fatal Error Global Event label: '%s'\n", value);

            if(pInfo->variables->IsEmpty(value))
            {
               ivLog->Print("         Fatal label undefined!\n");
               while(PopProcedure(pInfo) == IVR_SUCCESS)
               {
                  // Discard results, simply un-wind procedure calls so we can exit safely
               }
               bContinue = false;
               currentStep = NULL;
            }
            else
            {
               currentStep = currentTAM->GetLabelStep(value);

               // If IVR_FATAL didn't find a step in current TAM, pop the call
               // stack looking for the label.
               while(bContinue && (currentStep == NULL))
               {
                  if(PopProcedure(pInfo) == IVR_SUCCESS)
                  {
                     currentStep = currentTAM->GetLabelStep(value);
                  }
                  else
                  {
                     ivLog->Print("            Could not locate Fatal Error Global Event label: '%s'\n", value);
                     bContinue = false;
                  }
               }

               pInfo->variables->SetSystem("@FATAL", "");
            }

            delete[] value;
         }
         else
         {
            bContinue = false;
            currentStep = NULL;
         }

         if(currentStep == NULL)
         {
            status = IVR_FATAL;
			}
      }
      else if((currentStep == NULL) && (status == IVR_ENDPROC) && (pInfo->appFlags.status == lsNormal))
      {
 			pInfo->appFlags.status = lsNotRunning;
      }

      // Process any system events
      while(WaitForSingleObject(pInfo->messages->GetMessageSignal(), 0) == WAIT_OBJECT_0)
      {
         IV_STATUS status2 = pInfo->ivBoard->ProcessSystemEvent();
         if(status2 != IVR_SUCCESS)
         {
            bContinue = false;
            status = status2;
         }
      }
      // Check for shutdown after each step.
      if(pInfo->shutdownRequested && pInfo->safeToShutdown)
      {
         bContinue = false;
         status = IVR_SHUTDOWN;
      }
   }

   // pInfo->status will be lsCrashed if we have crashed even if
   // status is currently good because we crashed nicely such as the FATAL handler
   // Always make a shutdown normal
   if(((pInfo->appFlags.status != lsCrashed) && ((status == IVR_SUCCESS)
     	|| (status == IVR_ENDPROC))) || (status == IVR_SHUTDOWN))
	{
      pInfo->appFlags.status = lsNotRunning;
   }
   else
   {
      pInfo->appFlags.status = lsCrashed;
   }

   SendSimulatorData(pInfo);

   return(status);
}
/*TAP::Execute*/

/* -------------------------------------------------------------------------
   Name:        TAP::FirstProcName

   Purpose:     Get the first procedure (.TAM) file name from the
                application (.TAP) file.

   Notes:       - I don't like the fact that I have to allocate memory for
                  pszFirstProcName.

   Modified:    01-Aug-1996 Tim Forner: Application files now ".TAP" and
                            module files are now ".TAM".
                19-May-1997 Tim Forner: Re-written to handle new Designer
                            files.
   ------------------------------------------------------------------------- */
IV_STATUS TAP::FirstProcName()
{
   FILE      *fApp;
   IV_STATUS  status;
   int        pos;
   bool       cont;
// char      *s;
   int        counter;
   MYCHAR     s;

   memset (&s, 0, sizeof (MYCHAR));

   if (IvOpenFile (ivLog, &fApp, appName.arr) != IVR_SUCCESS)
      return IVR_FAIL;

   TxtSkipSection (fApp);                  // Skip past intro/notes
   TxtSkipSection (fApp);                  // Skip past header
   cont = true;
   status = IVR_SUCCESS;
   while (!feof (fApp) && cont) {
      cont = false;
      pos = TxtReadString (fApp, &s);       // Read 1st field.
      if (pos >= 0) {                       // Errors encountered?
         // Is it the "STaRT"ication record?
         if (!strcmp (s.arr, STRT)) {       // Yes. 14th fld is 1st proc. name.
            for (counter = 2; counter < 4; ++counter)
               TxtReadString (fApp, &s);    // Skip label and current step no.
            pos = TxtReadString (fApp, &s); // Read in the next step no.
            sscanf (s.arr, "%d", &iFirstProcID);
            if (pos >= 0)                   // NO Errors encountered?
               status = IVR_SUCCESS;        // Yes.  Successful.
         } else {                           // No. ...
            status = TxtNextRecord (fApp);  // ... Discard this record.
            if (status == IVR_SUCCESS) {    // Did we find end of record?
               status = IVR_FAIL;       // Yes. Reset 'rc' to fail (not found).
               cont = true;
            }
         }
      }
   }

   if (status != IVR_SUCCESS)
      ivLog->Print("   ERROR: Could not retrieve first procedure name/file from application\n");

   // Make sure application file is closed after loading - TF May 28/96
   if (IvCloseFile (ivLog, &fApp, appName.arr) != IVR_SUCCESS) {
      delete [] s.arr;
      return IVR_FAIL;
   }
   delete [] s.arr;
   return  status;
}
/*TAP::FirstProcName*/

/* -------------------------------------------------------------------------
   Name:        TAP::FileExist

   Notes:       - This code was using 'findfirst' but this function is being
                  reported as having a memory leak.  I replaced it with
                  'access'.
   ------------------------------------------------------------------------- */
bool TAP::FileExist (const char *path)
{
   return (access (path, 0) == 0);
}
/*TAP::FileExist*/

/* -------------------------------------------------------------------------
   Name:        TAP::SetMonData
   ------------------------------------------------------------------------- */
void TAP::SetMonData(PIVRINFO pInfo)
{
	const char* fileName;
   CLIENT_MSG appInfo;
   EXPANDED_CLIENT_MSG fullInfo;

   appInfo.appID = pInfo->iAppID;
   // Do not copy more than allowable TELEFLOW_APP_NAME_MAX characters as defined in ivr_defs.h
   strncpy(appInfo.appNm, pInfo->pszAppNm == NULL ? "" : pInfo->pszAppNm, TELEFLOW_APP_NAME_MAX);
   appInfo.appNm[TELEFLOW_APP_NAME_MAX] = '\0';
   if(pInfo->pszAppFile == NULL)
   {
      strcpy(appInfo.appFile, "");
   }
   else if((fileName = strrchr(pInfo->pszAppFile, '\\')) == NULL)
   {
   	// Do not copy more than allowable TELEFLOW_TAP_NAME_MAX characters as defined in ivr_defs.h
      strncpy(appInfo.appFile, pInfo->pszAppFile, TELEFLOW_TAP_NAME_MAX);
   }
   else
   {
   	// Do not copy more than allowable TELEFLOW_TAP_NAME_MAX characters as defined in ivr_defs.h
      strncpy(appInfo.appFile, fileName + 1, TELEFLOW_TAP_NAME_MAX);
   }
 	// Make sure string is null-terminated
   appInfo.appFile[TELEFLOW_TAP_NAME_MAX] = '\0';
   appInfo.lineNo = pInfo->iPort > 0 ? pInfo->iPort : -1;
   appInfo.stepNo = currentStep == NULL ? 0 : currentStep->GetCurrentStep();
   strncpy(appInfo.currentTAMName, currentTAM == NULL ? "" : currentTAM->GetShortName(), TELEFLOW_TAM_NAME_MAX);
   appInfo.currentTAMName[TELEFLOW_TAM_NAME_MAX] = '\0';
   strncpy(appInfo.stepName, currentStep == NULL ? "" : currentStep->GetStepName(), TELEFLOW_STEP_NAME_MAX);
   appInfo.stepName[TELEFLOW_STEP_NAME_MAX] = '\0';
   strncpy(appInfo.stepData, pInfo->variables->Get("@SYS_DATA") == NULL ? "" : pInfo->variables->Get("@SYS_DATA"), TELEFLOW_SYS_INFO_MAX);
   appInfo.stepData[TELEFLOW_SYS_INFO_MAX] = '\0';
   appInfo.status = pInfo->appFlags.status;
   appInfo.license = pInfo->portLicense;
   appInfo.restartCnt = pInfo->restartCnt;
   appInfo.restartMax = pInfo->bRestartInfinite ? -1 : pInfo->restartMax;
   appInfo.ivLog = pInfo->ivLog;
   //
   fullInfo.client_msg = appInfo;
   // Called number (DNIS) and calling number (ANI or CLID) available to Monitor
   strncpy(fullInfo.calledNumber,  (pInfo->calledNumber  == NULL) ? "" : pInfo->calledNumber,  MAX_DIGIT_BUFFER_SIZE);
   fullInfo.calledNumber[MAX_DIGIT_BUFFER_SIZE] = '\0';
   strncpy(fullInfo.callingNumber, (pInfo->callingNumber == NULL) ? "" : pInfo->callingNumber, MAX_DIGIT_BUFFER_SIZE);
   fullInfo.callingNumber[MAX_DIGIT_BUFFER_SIZE] = '\0';

   pInfo->tcpClientManager->UpdateAppInfo(fullInfo);
}

/*TAP::SetMonData*/

void TAP::SendSimulatorData(PIVRINFO pInfo)
{
   if(pInfo->hSimulator != NULL)
	{
      TFSimulatorCommunicationPacket* message = new TFSimulatorCommunicationPacket(TFSimulatorCommunicationPacket::TF_SIMULATOR_REFRESH, pInfo->iAppID);
      message->status = pInfo->appFlags.status;

      if(currentStep != NULL)
      {
         message->stepNumber = currentStep->GetCurrentStep();
         message->stepName = new char[strlen(currentStep->GetStepName()) + 1];
         strcpy(message->stepName, currentStep->GetStepName());
         message->TAMName = new char[strlen(currentTAM->GetShortName()) + 1];
         strcpy(message->TAMName, currentTAM->GetShortName());
      }

      pInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SIMULATOR, message);
   }
}

