/* -------------------------------------------------------------------------
   ------------------------------------------------------------------------- */
#ifndef TAP_H
#define TAP_H

#include "tam.h"
#include "txtfile.h"

#ifdef __DLL__
#define TAPLIBAPI __declspec(dllexport)
#else
#define TAPLIBAPI __declspec(dllimport)
#endif

/* -------------------------------------------------------------------------
   Class:       ArgLink

   Purpose:     This class stores the name of the argument and the name of
                the parameter that will accept the arguments value.

                When the called procedure exits the argument may be updated
                with the value of the parameter.

   Definitions: ARGUMENT:  variable / value passed into the procedure.
                PARAMETER: procedure variable that accepts value passed in
                           from calling procedure.

   Notes:       - When building the argument / parameter list, add the
                  parameters to the list first.  If there are more arguments
                  than parameters, the extra arguments are ignored.  If
                  there are less arguments than parameters, the remaining
                  parameters will be created and defaulted to a blank
                  string.
   ------------------------------------------------------------------------- */
class ArgLink : public Link
{
   public:
      ArgLink();
      ~ArgLink();

      void SetArgName (const char *value, int len = -1);
      void SetParmName (const char *value, int len = -1);
      void SetUpdateOnReturn (bool value) { bUpdateOnReturn = value; }
      const char *GetArgName() const { return pszArgNm; }
      const char *GetParmName() const { return pszParmNm; }
      bool GetUpdateOnReturn() const { return bUpdateOnReturn; }

   private:
      char  *pszArgNm;
      char  *pszParmNm;
      bool   bUpdateOnReturn;
      void   SetString (char **dest, const char *src, int len);
};

/* -------------------------------------------------------------------------
   Class:       QueueLink

   Purpose:     This class stores information about the current location in
                the current TAM before loading in the next TAM.  It is used
                to restore the current TAM when the next TAM exits.
   ------------------------------------------------------------------------- */
class QueueLink : public Link
{
   public:
      QueueLink();
      ~QueueLink();

      void SetTAM (TAM *pushTAM) { pTAM = pushTAM; }
      void SetStep (Step *pushStep) { pStep = pushStep; }
      void SetArgList (Head *pushList) {pArgList = pushList; }

      Step *GetStep() const { return pStep; }
      TAM  *GetTAM() const { return pTAM; }
      Head *GetArgList() const { return pArgList; }

   private:
      TAM      *pTAM;        // TAM to load after returning.
      Step     *pStep;       // Step to load and execute after returning.
      Head     *pArgList;    // Link list of parameters.
};

/* -------------------------------------------------------------------------
   Class:       TAP

   Notes:       - This class needs to be exported to allow the WAVEngine
                  access.
   ------------------------------------------------------------------------- */
class TAPLIBAPI TAP
{
   public:
      TAP(PIVRINFO pInfo);
      ~TAP(void);

      IV_STATUS Execute(PIVRINFO pInfo);
      bool GetReady(void);
      void SetSysVariables(PIVRINFO pInfo);
      IV_STATUS LoadParameters(PIVRINFO pInfo);
      IV_STATUS SetReferenceParameters(PIVRINFO pInfo, PIVRINFO parentInfo, const char* parameters);

   private:
      Log              *ivLog;
      Head             *TAMs;
      Step             *currentStep;
      TAM              *currentTAM;
      MYCHAR            appName;
      int               iFirstProcID;
      Head             *qList;
      bool              bReady;
      Head*             startParametersList;

      IV_STATUS LoadVoiceScripts(PIVRINFO pInfo);
      IV_STATUS LoadTAMs(PIVRINFO pInfo);
      IV_STATUS FirstProcName(void);
      TAM* FindTAMID(int id);
      TAM* FindTAM(const char *name);
      bool FileExist(const char *path);
      IV_STATUS PushProcedure(PIVRINFO pInfo);
      IV_STATUS PopProcedure(PIVRINFO pInfo);
      void SetMonData(PIVRINFO pInfo);
      void SendSimulatorData(PIVRINFO pInfo);
};

#undef TAPLIBAPI

#endif
