/* -------------------------------------------------------------------------
   ------------------------------------------------------------------------- */
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include "txtfile.h"
#include "..\escutil\iv_str.h"

/* -------------------------------------------------------------------------
   Name:        IvOpenFile
   ------------------------------------------------------------------------- */
IV_STATUS IvOpenFile (Log *ivLog, FILE **fp, const char *filename)
{
   *fp = fopen (filename, "rb");
   if (! *fp) {
      ivLog->Print ("\nERROR: Could not open file '%s'.\n", filename);
      *fp = 0;
      return IVR_FAIL;
   }

   return IVR_SUCCESS;
}
/*IvOpenFile*/


/* -------------------------------------------------------------------------
   Function:    IvCloseFile

   Purpose:     Closes the file pointed to by file handle and traps for any
                errors.
   ------------------------------------------------------------------------- */
IV_STATUS IvCloseFile (Log *ivLog, FILE **fp, const char *filename)
{
   if (fclose (*fp)) {
      ivLog->Print ("\nERROR: Could not close file '%s'.\n", filename);
      *fp = 0;
      return IVR_FAIL;
   }
   *fp = 0;
   return IVR_SUCCESS;
}
/*IvCloseFile*/


/* -------------------------------------------------------------------------
   Function:    TxtFetchStep

   Purpose:     Read all the fields of the current step record into a step
                structure.

   Returns:     IVR_SUCCESS
                IVR_FAIL

   Modified:    01-Aug-1996 (TF): Documented.
   ------------------------------------------------------------------------- */
IV_STATUS TxtFetchStep (
   FILE     *fp,        // File handle to TAM (procedure) to read.
   PSTEPREC  stepRec)   // Step structure to fill with step data.
{
                                                           // Record Fields:
   if (TxtReadNumber (fp, &(stepRec->op)) < 0)             //  1. Op code
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->label)) < 0)          //  2. Label
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->current)) < 0)        //  3. Current step
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->next)) < 0)           //  4. Next step
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->parent)) < 0)         //  5. Parent step
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u1)) < 0)             //  6. u1
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u2)) < 0)             //  7. u2
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u3)) < 0)             //  8. u3
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u4)) < 0)             //  9. u4
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u5)) < 0)             // 10. u5
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u6)) < 0)             // 11. u6
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u7)) < 0)             // 12. u7
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->u8)) < 0)             // 13. u8
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s1)) < 0)             // 14. s1
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s2)) < 0)             // 15. s2
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s3)) < 0)             // 16. s3
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s4)) < 0)             // 17. s4
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s5)) < 0)             // 18. s5
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s6)) < 0)             // 19. s6
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s7)) < 0)             // 20. s7
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s8)) < 0)             // 21. s8
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s9)) < 0)             // 22. s9
      return IVR_FAIL;

   if (TxtReadString (fp, &(stepRec->s10)) < 0)            // 23. s10
      return IVR_FAIL;

   if (TxtReadNumber (fp, &(stepRec->attrib)) < 0)         // 24. Attribute
      return IVR_FAIL;

   // Ignore remaining fields, they are not required by the Engine.
   //    25. Notes     27. Y         29. Tack Y          31. Parent Connect
   //    26. X         28. Tack X    30. Connect Shape

   // Read until end of record.
   return TxtNextRecord (fp);
}
/*TxtFetchStep*/


/* -------------------------------------------------------------------------
   Name:        TxtNextRecord

   Purpose:     Reads characters from file from current location to
                END_RECORD character.  END_RECORD character discarded.

   Returns:     IVR_SUCCESS - File pointer moved to beginning of next
                              record.
                IVR_FAIL    - End of file encountered.

   Notes:       - Used only to skip ahead to the start of the next
                  record (?)
   ------------------------------------------------------------------------- */
IV_STATUS TxtNextRecord (FILE *fp)
{
   int c = '\0';

   while (!feof (fp) && c != END_RECORD)   // Read until end of file or record.
      c = fgetc (fp);

   if (feof (fp))                          // Any more records?
      return IVR_FAIL;                     // No.  Reached end of file.

   return IVR_SUCCESS;                     // Yes. At beginning of next record.
}
/*TxtNextRecord*/


/* -------------------------------------------------------------------------
   Name:        TxtSkipSection

   Purpose:     Reads and discards characters from file starting at current
                location up to and including END_SECTION (^Z) character.

   Returns:     IVR_SUCCESS - File pointer moved to beginning of next
                              section.
                IVR_FAIL    - End of file encountered.

   Notes:       - File *must* be opened in binary mode
                - Used only to skip ahead to the start of the next
                  section.
   ------------------------------------------------------------------------- */
IV_STATUS TxtSkipSection (FILE *fp)
{
   int  c = '\0';

   while (!feof (fp) && c != END_SECTION)  // Read until EOF or END_SECTION.
      c = fgetc (fp);

   if (c == END_SECTION)
      return IVR_SUCCESS;                  // Yes. At beginning of next record.

   return IVR_FAIL;                        // No.  Reached end of file.
}
/*TxtSkipSection*/


/* -------------------------------------------------------------------------
   Name:        TxtReadNumber

   Purpose:     Reads in characters from the current field and converts the
                characters into a number.
   ------------------------------------------------------------------------- */
int TxtReadNumber (FILE *fp, int *num)
{
   char s[24];
   int  pos = 0;
   int  c;

   c = fgetc (fp);
   while (!feof (fp) && c != END_FIELD && c != END_RECORD && c != END_SECTION) {
      s[pos] = (char) c;
      pos++;
      c = fgetc (fp);
   }
   s[pos] = '\0';
   *num = atoi (s);

   switch (c) {
      case END_FIELD:     // End of field reached.  Return field length.
         return pos;

      case END_SECTION:   // End of section. If data, return field length...
         if (pos > 0)
            return pos;
         return -2;       // ... otherwise return end of section.

      case END_RECORD:    // End of record reached.
         return -2;

      default:
         return -1;       // End of file reached.
   }
}
/*TxtReadNumber*/

/* -------------------------------------------------------------------------
   Name:        TxtSetupStepRecMemory
   ------------------------------------------------------------------------- */
void TxtSetupStepRecMemory (PSTEPREC stepRec)
{
   memset (stepRec, 0, sizeof (STEPREC));
   stradd (stepRec->label.arr, "", stepRec->label.len);
   stradd (stepRec->s1.arr,    "", stepRec->s1.len);
   stradd (stepRec->s2.arr,    "", stepRec->s2.len);
   stradd (stepRec->s3.arr,    "", stepRec->s3.len);
   stradd (stepRec->s4.arr,    "", stepRec->s4.len);
   stradd (stepRec->s5.arr,    "", stepRec->s5.len);
   stradd (stepRec->s6.arr,    "", stepRec->s6.len);
   stradd (stepRec->s7.arr,    "", stepRec->s7.len);
   stradd (stepRec->s8.arr,    "", stepRec->s8.len);
   stradd (stepRec->s9.arr,    "", stepRec->s9.len);
   stradd (stepRec->s10.arr,   "", stepRec->s10.len);
}
/*TxtSetupStepRecMemory*/

/* -------------------------------------------------------------------------
   Name:        TxtFreeStepRecMemory
   ------------------------------------------------------------------------- */
void TxtFreeStepRecMemory (PSTEPREC stepRec)
{
   delete [] stepRec->label.arr;
   delete [] stepRec->s1.arr;
   delete [] stepRec->s2.arr;
   delete [] stepRec->s3.arr;
   delete [] stepRec->s4.arr;
   delete [] stepRec->s5.arr;
   delete [] stepRec->s6.arr;
   delete [] stepRec->s7.arr;
   delete [] stepRec->s8.arr;
   delete [] stepRec->s9.arr;
   delete [] stepRec->s10.arr;
   memset (stepRec, 0, sizeof (STEPREC));
}
/*TxtFreeStepRecMemory*/

/* -------------------------------------------------------------------------
   Name:        TxtReadString

   Purpose:     Reads in characters from the current field and stores them
                in a MYCHAR structure.

   Notes:       - Memory is automatically allocated to the MYCHAR structure
                  in 64 byte chunks.  Memory must be freed when the variable
                  is no longer needed.
   ------------------------------------------------------------------------- */
int TxtReadString (FILE *fp, MYCHAR *s)
{
   int     c;
   int     pos = 0;

   strclear (s->arr);
   c = fgetc (fp);
   while (!feof (fp) && c != END_FIELD && c != END_RECORD && c != END_SECTION) {
      charadd (s->arr, (char)c, s->len);
      c = fgetc (fp);
      pos++;
   }

   switch (c) {
      case END_FIELD:     // End of field reached.  Return field length.
         return pos;

      case END_SECTION:   // End of section. If data, return field length...
         if (pos > 0)
            return pos;
         return -2;       // ... otherwise return end of section.

      case END_RECORD:    // End of record reached.
         return -2;

      default:
         return -1;       // End of file reached.
   }
}
/*TxtReadString*/

/* -------------------------------------------------------------------------
   Name:        TxtReadField

   Purpose:     Reads characters from file from current location to
                END_FIELD character.  END_FIELD character discarded.

   Returns:     -1    End of file reached.
                -2    End of record reached.
                 N    Length of field (N >= 0).

   Notes:       - Parameter 's' must be large enough to hold the
                  expected field; no checks against string overflows
                  are performed.
                - END_FIELD, END_SECTION, etc. characters are discarded.
   ------------------------------------------------------------------------- */
int TxtReadField (
   FILE *fp,                 // File to read.
   char *s)                  // String to store field value.
{
   int   pos = 0;
   int   c;

   c = fgetc (fp);
   while (!feof (fp) && c != END_FIELD && c != END_RECORD && c != END_SECTION) {
      s[pos] = (char) c;
      ++pos;
      c = fgetc (fp);
   }
   s[pos] = '\0';

   switch (c) {
      case END_FIELD:     // End of field reached.  Return field length.
         return pos;

      case END_SECTION:   // End of section. If data, return field length...
         if (pos > 0)
            return pos;
         return -2;       // ... otherwise return end of section.

      case END_RECORD:    // End of record reached.
         return -2;

      default:
         return -1;       // End of file reached.
   }
}
/*TxtReadField*/


/* -------------------------------------------------------------------------
   Name:        TxtApplHeader

   Purpose:     Read the header information from the application (.TAP)
                file.

   Returns:     IVR_SUCCESS    Application header read successfully
                IVR_FAIL       Bad header and/or version number
                IVR_FATAL      Could not open/read app. file

   Notes:       - In *.TAP's header record is as follows:

                  ENGTF@TAP 0200 00000 __________ O
                  0........ 9... 1.... 1...  ...6 6
                                 3     7        2 3

                  0-4      "ENGTF" a constant that defines this as
                                    an Engine Teleflow file
                    5      "@"     a separator character
                  6-8      "TAP"   defines this file's type
                                    ("Telephone APplication")

                  9-12     Engine/Designer version file was built to run
                           with

                  13-16    Version/serial number of this particular file

                  18-62    Unused space

                  63       End of section marker (^Z)

   Modified:    07-Aug-96 (TF): Created.
   ------------------------------------------------------------------------- */
IV_STATUS TxtApplHeader (Log *ivLog, const char *appDb)
{
   FILE      *fApp;
   IV_STATUS  rc;

   rc = IvOpenFile (ivLog, &fApp, appDb);
   if (rc != IVR_SUCCESS)
      return rc;

   rc = TxtGenericHeader (ivLog, fApp, APP_TYPE);

   // Make sure application file is closed after loading
   if (IvCloseFile (ivLog, &fApp, appDb) != IVR_SUCCESS)
      rc = IVR_FAIL;

   return rc;
}
/*TxtApplHeader*/


/* -------------------------------------------------------------------------
   Function:    TxtModuleHeader

   Purpose:     Read the header information from the module (.TAM) file.

   Returns:     IVR_SUCCESS    Module has a header and it was read
                IVR_FAIL       Module does not have a header
                IVR_FATAL      Could not read from module file

   Notes:       - Old module files may not have a header.
                - Currently, this module ignores a "header not found" error

                - In *.TAM's first 64 characters are:

                  ENGTF@TAM 0200 00000 __________ O O
                  1........ 1... 1.... 1...  ...6 6 6
                            0    4     8        2 3 4

                  1-9      "ENGTF" a constant that defines this as
                                    an Engine Teleflow file
                           "@"     a separator character
                           "TAM"   defines this file's type
                                    ("Telephone Application Module")

                  10-13    Engine/Designer this file is built to run with

                  14-17    Version of this particular file, as an
                           incrementing number

                  18-62    Unused space

                  63       End of field marker (001)

                  64       End of record marker (002)

   Modified:    07-Aug-96 (TF): Created.
   ------------------------------------------------------------------------- */
IV_STATUS TxtModuleHeader (
   FILE *fproc)               // File pointer of module file.
{
   char       s[255];
   IV_STATUS  rc;


   if (TxtReadField (fproc, s) < 0)       // Read header field
      return IVR_FATAL;                   // Errors encountered.

   if (strncmp (s, "ENGTF", 5) == 0) {    // Is it an "ENGTF" type
      rc = IVR_SUCCESS;                   // This is the header
      TxtNextRecord (fproc);              // Read until end of record.

   } else {
      rc = IVR_FAIL;                      // Module file does not have a header
      fseek (fproc, 0, SEEK_SET);         // "Rewind" to beginning of file
   }

   return rc;
}
/*TxtModuleHeader*/


/* -------------------------------------------------------------------------
   Name:        TxtGenericHeader

   Purpose:     Read header information from "generic" header structure
                of an open Teleflow file.

   Returns:     IVR_SUCCESS    Application header read successfully
                IVR_FAIL       Bad header and/or version number
                IVR_FATAL      Could not open/read app. file

   Notes:       - Header record is as follows:

                  ENGTF@TAP 0200 00000 __________ O
                  0........ 9... 1.... 1...  ...6 6
                                 3     7        2 3

                  0-4      "ENGTF" a constant that defines this as
                                    an Engine Teleflow file
                    5      "@"     a separator character
                  6-8      "TAP"   defines this file's type
                                    ("Telephone APplication")

                  9-12     Engine/Designer version file was built to run
                           with

                  13-16    Version/serial number of this particular file

                  18-62    Unused space

                  63       End of section marker (^Z)

   Modified:   22-May-97 (TF): Created.
   ------------------------------------------------------------------------- */
IV_STATUS TxtGenericHeader (
   Log  *ivLog,
   FILE *fproc,              // File pointer of an open TeleFlow file.
   char *ftype)              // Type of file (e.g. "TAP" or "TAM").
{
   char        s[255];
   int         pos;
   IV_STATUS   rc;

   char  filecreate[6];      // File creation identifier (ie: "ENGTF")
   char  filetype[4];        // TAP, TAM, etc.
   char  filevers[5];        // Version of Designer that file was built with
   char  fileserial[6];      // Serial number of this file

   // Skip past Introduction/Notes section
   TxtSkipSection (fproc);

   // Read header...
   pos = TxtReadField (fproc, s);          // Read 1st field.

   // Break Header into its components
   strncpy (filecreate, s, 5);
   filecreate[5] = '\0';
   strncpy (filetype, s + 6, 3);
   filetype[3] = '\0';
   strncpy (filevers, s + 9, 4);
   filevers[4] = '\0';
   strncpy (fileserial, s + 13, 5);
   fileserial[5] = '\0';

   // Check for errors...
   if (pos < 0) {
      ivLog->Print ("\nERROR: Could not read file.\n");
      rc = IVR_FATAL;

   } else if (strcmp (filecreate, "ENGTF")) {
      ivLog->Print ("\nERROR: File is not an Teleflow file.\n");
      rc = IVR_FAIL;

   } else if (strcmp (filetype, ftype)) {
      ivLog->Print ("\nERROR: File is not of type '%s'.\n", ftype);
      rc = IVR_FAIL;

   } else if (strcmp (filevers, DESIGNVER)) {
      ivLog->Print ("\nERROR: File is not of the expected version.\n"
                   "Use TeleFlow Designer to convert.\n");
      rc = IVR_FAIL;

   } else
      rc = IVR_SUCCESS;       // No errors!

   if (ivLog->GetDebug ('p')) {
#if 0
      if (strcmp (filetype, APP_TYPE) == 0)
         ivLog->Print (" Engenic TeleFlow Application");

      else if (strcmp (filetype, MOD_TYPE) == 0)
         ivLog->Print (" Engenic TeleFlow Module");

      else
         ivLog->Print (" Engenic TeleFlow File");
#endif
      ivLog->Print (" Ver: %s S/N: %s\n", filevers, fileserial);
   }

   return rc;
}
/*TxtGenericHeader*/

/* -------------------------------------------------------------------------
   Name:        DefaultExt

   Purpose:     Appends a default extenstion to a file/path specification
                if that specification does not include an extension.

                If the extension needs to be added to the path variable, the
                memory allocated to 'path' will be freed and re-allocated.

   Modified:    23-May-1997 Tim Forner: Created.
                26-Feb-1998 Jamie Royer: Changed logic to dynamically
                   manage memory.  The return logic changed to return false
                   only if path is not defined.
   ------------------------------------------------------------------------- */
bool DefaultExt (MYCHAR *path, const char *ext)
{
   char  ldrive[MAXDRIVE];
   char  ldir[MAXDIR];
   char  lname[MAXFILE];
   char  lexten[MAXEXT];

   if (!path || !path->arr || !*path->arr)
      return false;                               // Invalid pointer passed in.

   if (!ext || !*ext)
      return true;                  // No extension passed in, path is correct.

   fnsplit (path->arr, ldrive, ldir, lname, lexten);

   if (!*lexten) {
      if (!strstr (ext, "."))
         charadd (path->arr, '.', path->len);
      stradd (path->arr, ext, path->len);
   }

   return true;                                             // Path is correct.
}
/*DefaultExt*/

/* -------------------------------------------------------------------------
   Function:    ChangePath

   Purpose:     Change working directory to location of the application
                file.

   Returns:     FALSE    No path in 'fullName' parameter, working directory
                           and the 'pathOnly' parameter remain unchanged.
                TRUE     Path found in file name, working directory has
                           been changed and directory changed to specified
                           in 'pathOnly' parameter

   Modified:    21-Jan-1998 Tim Forner: Created.
                26-Feb-1998 Jamie Royer: Changed logic to dynamically
                   manage memory.  The return logic changed to return false
                   only if path variable is not defined or path doesn't
                   exist.
   ------------------------------------------------------------------------- */
bool ChangePath (char *path)
{
   char  ldrive[MAXDRIVE];
   char  ldir[MAXDIR];
   char  lname[MAXFILE];
   char  lexten[MAXEXT];
// char *newPath;
   int   result;

   if (!path || !*path)
      return false;                               // Invalid pointer passed in.

   fnsplit (path, ldrive, ldir, lname, lexten);

   if (*ldrive && *ldir) {
      result = chdir (ldrive);
      if (!result)
         result = chdir (ldir);
/*
      newPath = new char [strlen (ldrive) + strlen (ldir) + 1];
      strcpy (newPath, ldrive);
      strcat (newPath, ldir);
      result = chdir (newPath);
      delete [] newPath;
*/

   } else if (*ldrive)
      result = chdir (ldrive);

   else if (*ldir)
      result = chdir (ldir);

   else
      result = true;                 // We didn't change directory, no problem.

   if (result)
      return false;

   return true;
}
/*ChangePath*/

