/* -------------------------------------------------------------------------
   File:        LOG.CPP

                Copyright (c) 1997-2008 Engenic
                All rights reserved.
   ------------------------------------------------------------------------- */
#include <stdlib.h>
#include <stdarg.h>
#include <iostream.h>
#include <string.h>
#include <sys/stat.h>
#include "ivr_defs.h"
#include "log.h"

/* -------------------------------------------------------------------------
   Name:        Log::Log                                       (Constructor)
   ------------------------------------------------------------------------- */
Log::Log(int maxFileSize)
{
   fp            = NULL;
   toScreen      = false;
   bAtLeftMargin = true;
   bTimeStamp    = true;
   iMaxFileSize = maxFileSize;
   disabled = false;
   
   memset(debug, 0, sizeof(bool) * 256);

   // Allows 'true' and 'false' to be passed as parameters to 'Debug'.
   SetDebug(1);
}
/*Log::Log*/

/* -------------------------------------------------------------------------
   Name:        Log::~Log                                       (Destructor)
   ------------------------------------------------------------------------- */
Log::~Log(void)
{
   Close();
}
/*Log::~Log*/

/* -------------------------------------------------------------------------
   Name:        Log::Open

   Notes:       - Data appended to end of file.
                - If this function is unable to open the file, "fp" will be
                  NULL.
   ------------------------------------------------------------------------- */
void Log::Open(const char* s, bool bAppend, bool bUnbuffered)
{
   Close();

   bFileAppend     = bAppend;
   bFileUnbuffered = bUnbuffered;

   if(s != NULL)
   {
      string fileName;

      currentFileTime.Update();

      fileName = logName = s;
      fileName += "_";
      fileName += currentFileTime.GetDate("-");
      fileName += ".log";

      if((fp = fopen(fileName.c_str(), bAppend ? "a" : "w")) != NULL)
      {
         if(bUnbuffered)
         {
            setbuf(fp, 0);
         }

         fputs("\nEngenic TeleFlow: ", fp);
         fputs(TFVERSION, fp);
         fputs("\n(c) 1994 - 2008 Engenic\nAll rights reserved.\n\n", fp);
      }
   }
   else
   {
      logName.clear();
   }
}
/*Log::Open*/

/* -------------------------------------------------------------------------
   Name:        Log::Close
   ------------------------------------------------------------------------- */
void Log::Close(void)
{
   if(fp != NULL)
   {
      fclose(fp);
      fp = NULL;
   }
}
/*Log::Close*/

/* -------------------------------------------------------------------------
   Name:        Log::EnableScreenOutput
   ------------------------------------------------------------------------- */
void Log::EnableScreenOutput (bool bEnable)
{
   toScreen = bEnable;
}
/*Log::EnableScreenOutput*/

/* -------------------------------------------------------------------------
   Name:        Log::SetDebug
   ------------------------------------------------------------------------- */
void Log::SetDebug (int pos)
{
   if((pos >= 0) && (pos <= 255))
   {
      debug[pos] = true;
   }
}
/*Log::SetDebug*/

/* -------------------------------------------------------------------------
   Name:        Log::ClearDebug
   ------------------------------------------------------------------------- */
void Log::ClearDebug(int pos)
{
   if((pos >= 0) && (pos <= 255))
   {
      debug[pos] = false;
   }
}
/*Log::ClearDebug*/


/* -------------------------------------------------------------------------
   Name:        Log::GetDebug
   ------------------------------------------------------------------------- */
bool Log::GetDebug(int pos) const
{
   if((pos >= 0) && (pos <= 255))
   {
      return(debug[pos]);
   }

   return(false);
}
/*Log::ClearDebug*/

/* -------------------------------------------------------------------------
   Name:        Log::SeparateLines
   ------------------------------------------------------------------------- */
void Log::SeparateLines(char* pszLine)
{
   //                       "YYYY/MM/DD HH:MM:SS.ms_:  "
   char* pszIndent = "                          ";
   char szPrefix[27];
   char* sol;
   char* eol;
   std::string as;

   szPrefix[0] = '\0';
   sol = pszLine;

   while((sol != NULL) && (*sol != '\0'))
   {
      if((eol = strchr(sol, '\n')) != NULL)
      {
         *eol = '\0';  // Temporarily replace newline character with a NUL.
      }

      // If the cursor is at the left margin, print the timestamp for the
      // first line.  Every subsequent line will be indented to go past
      // the date (so text line up).
      if(bTimeStamp && bAtLeftMargin)
      {
         if(sol == pszLine)
         {
            Stamp(szPrefix);
         }
         else
         {
            strcpy(szPrefix, pszIndent);
         }
      }

      as = szPrefix;

      // Call the derived class' PrintLine() Function.
      as += sol;

      PrintLine(as, eol ? true : false);

      // Don't allow the file size to exceed iMaxFileSize.
      if(fp != NULL)
      {
         struct stat statbuf;
         ESCTime currentTime;

         currentTime.Update();

         // We have rolled over to a new day, so close the current log file
         // and start a new one
         if(atoi(currentTime.GetDate("")) > atoi(currentFileTime.GetDate("")))
         {
            bool bAppending = bFileAppend;
            bool bNoBuffering = bFileUnbuffered;
            Open(logName.c_str(), bAppending, bNoBuffering);
         }

         fstat(_fileno (fp), &statbuf);

         if((iMaxFileSize == -1) || (statbuf.st_size / 1024 < iMaxFileSize))
         {
            fputs(szPrefix, fp);
            fputs(sol, fp);
            if(eol)
            {
               fputc('\n', fp);
            }
         }
      }

      if(eol != NULL)
      {
         *eol = '\n';           // Restored character.
         eol++;                 // Move past newline.
         sol = eol;             // Reset beginning of line.
         bAtLeftMargin = true;  // Starting in the left margin (after newline).
      }
      else
      {
         bAtLeftMargin = false; // No newline encountered.
         sol = NULL;               // End of line found.
      }
   }
}
/*Log::SeparateLines*/

/* -------------------------------------------------------------------------
   Name:        Log::Print

   Notes:       - Passing NULL as the first parameter to 'vsnprintf' to
                  determine the number of characters needed for the buffer
                  doesn' work with all compilers.  Borland returns the
                  number of character needed.  Microsoft returns an error.
   ------------------------------------------------------------------------- */
int Log::Print (const char *format, ...)
{
   va_list ap;
   char* pszBuf;
   int nBufSize;

   if((format == NULL) || (format[0] == '\0') || disabled)
   {
      return(0);
   }

   // Determine required buffer size by passing null as first parameter.
   // ** This may not work with all compilers!  It works with Borland.
   va_start(ap, format);
   nBufSize = vsnprintf(0, 0, format, ap);
   va_end(ap);

   // If there is something to print allocate memory and process
   // the format string again, this time with a buffer to hold the string.
   if(nBufSize > 0)
   {
      pszBuf = new char[nBufSize + 1];
      va_start(ap, format);
      vsnprintf(pszBuf, nBufSize + 1, format, ap);
      va_end(ap);

      SeparateLines(pszBuf);

      delete[] pszBuf;
   }

   return(nBufSize);                      // Return number of characters output.
}
/*Log::Print*/

/* -------------------------------------------------------------------------
   Name:        Log::Debug

   Notes:       - I can't think of a way to call "Print" so I duplicated the
                  code from "Print".
   ------------------------------------------------------------------------- */
int Log::Debug(int pos, const char* format, ...)
{
   va_list ap;
   char* pszBuf;
   int nBufSize;

   if((format == NULL) || (format[0] == '\0') || !debug[pos] || disabled)
   {
      return(0);
   }

   // Determine required buffer size by passing null as first parameter.
   // ** This may not work with all compilers!  It works with Borland.
   va_start(ap, format);
   nBufSize = vsnprintf(0, 0, format, ap);
   va_end(ap);

   // If there is something to print allocate memory and process
   // the format string again, this time with a buffer to hold the string.
   if(nBufSize > 0)
   {
      pszBuf = new char[nBufSize + 1];
      va_start(ap, format);
      vsnprintf(pszBuf, nBufSize + 1, format, ap);
      va_end(ap);

      SeparateLines(pszBuf);

      delete[] pszBuf;
   }
   
   return(nBufSize);                      // Return number of characters output.
}
/*Log::Debug*/

/* -------------------------------------------------------------------------
   Name:        Log::EnableTimeStamp
   ------------------------------------------------------------------------- */
void Log::EnableTimeStamp (bool bEnable)
{
   bTimeStamp = bEnable;
}
/*Log::EnableTimeStamp*/

/* -------------------------------------------------------------------------
   Name:        Log::Stamp                                         (private)

   Notes:       - 'pszDate' must be at least 21 characters.
                - Stamp format: 'Mon DD HH:MI:SS.MM: '
   ------------------------------------------------------------------------- */
void Log::Stamp(char* pszDate)
{
   ESCTime st;

   st.Update();

   sprintf(pszDate, "%s %s:  ", st.GetDate(), st.GetTime());
}
/*Log::Stamp*/

void Log::Disable(bool disable)
{
   disabled = disable;
}

