/* -------------------------------------------------------------------------
   Rules:       - Client requests data from the server.
                - Server acknowledges all messages.
                - All messages have a header possibly followed by a data
                  block.
                - If error occurs sending the header block, close the
                  connection.
                - If error occurs receiving data, close the connection.
                - If WSAEWOULDBLOCK error occurs sending data, wait for
                  wait for write event and continue sending.
                - No new messages will be sent or received until the current
                  message has been processed and sent.
                - If data write pending, reads are discarded.
                - If a full response is not received for a request after ten
                  seconds, close the connection.

   Notes:       - There can be multiple clients connected to the server.
                  The server will create a new thread to communicate with
                  each client independently.  This code needs to be able to
                  handle multiple client connections.
   ------------------------------------------------------------------------- */
#include <winsock2.h>    // BCB6 upgrade - TF Mar.28/02
#include <windows.h>
#include <process.h>
#include "server_mon.h"
#include "inet_msg.h"
#include "TFServer_communication.h"
#include "TFTCPClient_communication.h"
#include "..\TFEngine\TFEngine_communication.h"

typedef struct
{
   NET_STORAGE           p;
   HANDLE                hSDEvent;
   HANDLE                hNetEvent;
   CommunicationHub<TFServerCommunicationPacket::TF_DESTINATION_ID>* commHub;
   TFTCPClientCommunicationDestination* messages;
   IVRApps              *ivrApps;
   Log                  *ivLog;
   EXPANDED_CLIENT_MSG  *appInfo;
   int                   appInfoCnt;
   int appID;
   short internetPort;
} MON_THREAD_INFO;

typedef struct
{
   NET_MSG  msg;
   int      msgRemaining;
   void    *msgPtr;
   void    *clientData;
   int      dataRemaining;
   void    *dataPtr;
} NET_INFO;

/* -------------------------------------------------------------------------
   Class:       SendData
   ------------------------------------------------------------------------- */
class SendData : public Link
{
public:
   char  *buf;
   char  *sendPtr;
   int    remaining;

   SendData (void *data, int dataSize);
   ~SendData();
};

/* -------------------------------------------------------------------------
   Class:       MonThreadLink
   ------------------------------------------------------------------------- */
class MonThreadLink : public Link
{
public:
   MON_THREAD_INFO  tInfo;

   MonThreadLink() : Link()
   {
      tInfo.hSDEvent  = CreateEvent (0, false, false, 0);
      tInfo.hNetEvent = CreateEvent (0, false, false, 0);
   }
   ~MonThreadLink()
   {
      CloseHandle (tInfo.hSDEvent);
      CloseHandle (tInfo.hNetEvent);
   }
};

/* -------------------------------------------------------------------------
   ------------------------------------------------------------------------- */
static unsigned _stdcall ServerMonClient (void *data);
unsigned _stdcall ServerMonClient (void *data);
bool HandleRead (MON_THREAD_INFO *tInfo, Head *myQueue, NET_INFO *info);
bool HandleWrite (NET_STORAGE *p, Head *myQueue);
bool FlagChangedApplications (MON_THREAD_INFO *tInfo, bool bRefresh,
                              int &sendInfoCnt);
bool SendUpdate (MON_THREAD_INFO *tInfo, Head *myQueue, bool bRefresh);
bool SendOutput (MON_THREAD_INFO *tInfo, Head *myQueue, NET_INFO *info);
bool SendAppMessage(MON_THREAD_INFO* tInfo, int appID, TFEngineCommunicationPacket::TF_ENGINE_EVENT eventID);
bool BroadcastAppMessage(MON_THREAD_INFO* tInfo, TFEngineCommunicationPacket::TF_ENGINE_EVENT eventID);
bool SendAppStopStreamMessage(MON_THREAD_INFO* tInfo, int appID);
bool SendAppStartStreamMessage(MON_THREAD_INFO* tInfo, int appID, int telephonyPort);
bool SendAppPauseStreamMessage(MON_THREAD_INFO* tInfo, int appID);

/* -------------------------------------------------------------------------
   Name:        SendData::SendData
   ------------------------------------------------------------------------- */
SendData::SendData (void *data, int dataSize) : Link()
{
   buf = new char [dataSize];
   memcpy (buf, data, dataSize);
   sendPtr = buf;
   remaining = dataSize;
}
/*SendData::SendData*/

/* -------------------------------------------------------------------------
   Name:        SendData::~SendData
   ------------------------------------------------------------------------- */
SendData::~SendData()
{
   delete [] buf;
}
/*SendData::~SendData*/

/* -------------------------------------------------------------------------
   Name:        ServerMonitorMain

   Purpose:     Waits for Monitor clients to connect and spawns a thread to
                handle the communication.  Up to 5 worker threads can be
                spawned.

   Notes:       - Need to be able to send events to the service thread.
   ------------------------------------------------------------------------- */
unsigned _stdcall ServerMonitorMain (void *arg)
{
   Head                  tHead;
   MonThreadLink        *tLink;
   NET_STORAGE           p;
   int                   event;
   DWORD                 ret;
   unsigned              dwThreadID;
   HANDLE                hMonThread;
   HANDLE                waitObjects[3];
   SERVER_MONITOR_INFO  *sInfo = (SERVER_MONITOR_INFO*)arg;
   bool                  bShutdown = false;
   TFServerCommunicationPacket* srvrEvent = new TFServerCommunicationPacket(TFServerCommunicationPacket::SRV_EVT_MON_TERMINATED, 0);

   // Initialize NT socket library.
   if (!NetInitialize()) {
      NetShutdown();
      // Send a message to the primary thread before returning.
      sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
      return 0;
   }

   // Create a socket to listen for connections.
   if (!NetListen (&p, sInfo->port)) {
      NetShutdown();
      // Send a message to the primary thread before returning.
      sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
      return 0;
   }

   waitObjects[0] = sInfo->messages->GetMessageSignal();  // Messages from TFServer and TCPClients
   waitObjects[1] = p.hEvent;                     // Client connections.

   // Wait for events: clients connections, worker thread termination,
   // service shutdown.
   while(!bShutdown || !tHead.Empty())
   {
      ret = WaitForMultipleObjects(2, waitObjects, false, INFINITE);
      if (ret == WAIT_FAILED)
      {
         NetCloseSocket(&p);
         NetShutdown();
         // Send a message to the primary thread before returning.
         sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
         return 0;
      }

      // Message from TFServer or TCPClient indicated that it's finished
		if(ret == WAIT_OBJECT_0)
      {
         TFTCPClientCommunicationPacket* message = (TFTCPClientCommunicationPacket*) sInfo->messages->ReceiveMessage();

         if(message != NULL)
         {
   	      // Service has signalled for this thread to terminate.
            if(message->shutDownThread == 0)
            {
               for(tLink = (MonThreadLink*) tHead.First(); tLink != NULL;
                  tLink = (MonThreadLink*) tLink->Suc())
               {
                  SetEvent(tLink->tInfo.hSDEvent);
               }
            	bShutdown = true;
   	      }
            // Worker thread is sending an event to indicate that it is done.
            else
            {
               for(tLink = (MonThreadLink*) tHead.First();
                  (tLink != NULL) && (tLink->tInfo.p.hSocket != (SOCKET) message->shutDownThread);
                  tLink = (MonThreadLink*) tLink->Suc());

               if(tLink != NULL)
               {
                  NetCloseSocket(&(tLink->tInfo.p));
                  delete sInfo->messages->Remove(message->shutDownThread);
                  delete tLink;
               }
            }

            delete message;
         }
      }

      // A client (Monitor application) is connecting to the server.
      // Don't allow new connections if service requested shutdown.
      else if (!bShutdown && ret == WAIT_OBJECT_0 + 1)
      {
         if (!NetGetEvent (&p, event))
         {
            NetCloseSocket(&p);
            NetShutdown();
            // Send a message to the primary thread before returning.
            sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
            return 0;
         }

         tLink = new MonThreadLink;
         tLink->tInfo.commHub = sInfo->commHub;
         tLink->tInfo.ivrApps = sInfo->ivrApps;
         tLink->tInfo.ivLog = 0;
         if (!NetAccept (&p, &(tLink->tInfo.p), tLink->tInfo.hNetEvent))
         {
            NetCloseSocket (&(tLink->tInfo.p));
            NetShutdown();
            delete tLink;
            // Send a message to the primary thread before returning.
            sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
            return 0;
         }
         tLink->tInfo.messages = new TFTCPClientCommunicationDestination(tLink->tInfo.p.hSocket);
         tLink->tInfo.appID = 0;

         sInfo->messages->Add(tLink->tInfo.messages);
         tHead.AddLast((Link*) tLink);

         hMonThread = (HANDLE) _beginthreadex (
            NULL,               // pointer to thread security attributes
            0,                  // initial thread stack size, in bytes
            ServerMonClient,    // pointer to thread function
            &(tLink->tInfo),    // argument for new thread
            0,                  // creation flags
            &dwThreadID         // pointer to returned thread identifier
         );
         CloseHandle (hMonThread);
      }
   }

   NetCloseSocket(&p);
   NetShutdown();
   // Send a message to the primary thread before returning.
   sInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, srvrEvent);
   return 0;
}
/*ServerMonitorMain*/

/* -------------------------------------------------------------------------
   Name:        ServerMonClient

   Notes:       - Need a way to tell the ServerMonitorMain thread when this
                  thread terminates.
   ------------------------------------------------------------------------- */
unsigned _stdcall ServerMonClient (void *data)
{
   MON_THREAD_INFO* tInfo = (MON_THREAD_INFO*) data;
   int               event;
   bool shutdown = false;
   DWORD             ret;
   HANDLE            waitObjects[3];
   Head              myQueue;
   NET_INFO          info;
   int               changedCnt;
   TFTCPClientCommunicationPacket* message;

   memset (&info, 0, sizeof (NET_INFO));
   info.msgRemaining = sizeof(NET_MSG);
   info.msgPtr = &info.msg;

   if(data == NULL)
   {
      return 0;
   }

   // Build an initial list of all messages.
   tInfo->appInfo = 0;
   tInfo->appInfoCnt = 0;
   tInfo->appID = 0;
   FlagChangedApplications (tInfo, true, changedCnt);

   waitObjects[0] = tInfo->hSDEvent;
   waitObjects[1] = tInfo->hNetEvent;
   waitObjects[2] = tInfo->messages->GetMessageSignal();

   while(!shutdown)
   {
      ret = WaitForMultipleObjects(3, waitObjects, false, INFINITE);
      switch(ret)
      {
         case WAIT_FAILED:
         case WAIT_OBJECT_0:
         {
            shutdown = true;
            break;
         }

         // An event has arrived on the socket.
         case WAIT_OBJECT_0 + 1:
         {
            if(!NetGetEvent(&(tInfo->p), event))
            {
               shutdown = true;
               break;
            }

            switch(event)
            {
               case NET_EVT_CLOSED:
               {
// For debugging
// SOCKADDR_IN destination;
// int addressSize = sizeof(SOCKADDR_IN);
// getpeername(tInfo->p.hSocket, (sockaddr*) &destination, &addressSize);
// printf("%s:%d disconnected\n", inet_ntoa(destination.sin_addr), ntohs(destination.sin_port));
                  if(tInfo->appID != 0)
                  {
                     SendAppStopStreamMessage(tInfo, tInfo->appID);
                  }
                  shutdown = true;
                  break;
               }

               case NET_EVT_WRITE:
               {
                  if(!HandleWrite(&tInfo->p, &myQueue))
                  {
                     shutdown = true;
                  }
                  break;
               }

               case NET_EVT_READ:
               {
                  if(!HandleRead(tInfo, &myQueue, &info))
                  {
                     shutdown = true;
                  }
                  break;
               }
            }
            break;
         }

         // Message
         case WAIT_OBJECT_0 + 2:
         {
            message = (TFTCPClientCommunicationPacket*) tInfo->messages->ReceiveMessage();

            if(message != NULL)
            {
               switch(message->event)
               {
                  case TFTCPClientCommunicationPacket::TCPCLIENT_EVT_SHUTDOWN:
                  {
                     if(tInfo->appID != 0)
                     {
                        SendAppStopStreamMessage(tInfo, tInfo->appID);
                     }
                     shutdown = true;
                     break;
                  }

                  // Send start stream message
                  case TFTCPClientCommunicationPacket::TCPCLIENT_EVT_START_STREAM:
                  {
// For debugging
// SOCKADDR_IN destination;
// int addressSize = sizeof(SOCKADDR_IN);
// getpeername(tInfo->p.hSocket, (sockaddr*) &destination, &addressSize);
// printf("Sending (%d) start to %s:%d\n", message->format, inet_ntoa(destination.sin_addr), ntohs(destination.sin_port));

                     SendData* myData;
                     NET_MSG msg;

                     msg.id = NET_EVT_START_STREAM;
                     msg.cause = NET_CAUSE_NONE;
                     msg.value = 0;
                     msg.dataSize = 0;
                     msg.data = (void*) message->format;

                     myData = new SendData(&msg, sizeof(NET_MSG));
                     myQueue.AddLast((Link*) myData);
                     if(!HandleWrite(&tInfo->p, &myQueue))
                     {
                        shutdown = true;
                     }
                     break;
                  }

                  // Send stop stream message
                  case TFTCPClientCommunicationPacket::TCPCLIENT_EVT_STOP_STREAM:
                  {
// For debugging
// SOCKADDR_IN destination;
// int addressSize = sizeof(SOCKADDR_IN);
// getpeername(tInfo->p.hSocket, (sockaddr*) &destination, &addressSize);
// printf("Sending stop to %s:%d\n", inet_ntoa(destination.sin_addr), ntohs(destination.sin_port));

                     SendData* myData;
                     NET_MSG msg;

                     msg.id = NET_EVT_STOP_STREAM;
                     msg.cause = NET_CAUSE_NONE;
                     msg.value = 0;
                     msg.dataSize = 0;

                     myData = new SendData(&msg, sizeof(NET_MSG));
                     myQueue.AddLast((Link*) myData);
                     HandleWrite(&tInfo->p, &myQueue);
                     shutdown = true;
                     break;
                  }

                  // Send pause stream message
                  case TFTCPClientCommunicationPacket::TCPCLIENT_EVT_PAUSE_STREAM:
                  {
// For debugging
// SOCKADDR_IN destination;
// int addressSize = sizeof(SOCKADDR_IN);
// getpeername(tInfo->p.hSocket, (sockaddr*) &destination, &addressSize);
// printf("Sending pause to %s:%d\n", inet_ntoa(destination.sin_addr), ntohs(destination.sin_port));

                     SendData* myData;
                     NET_MSG msg;

                     msg.id = NET_EVT_PAUSE_STREAM;
                     msg.cause = NET_CAUSE_NONE;
                     msg.value = 0;
                     msg.dataSize = 0;

                     myData = new SendData(&msg, sizeof(NET_MSG));
                     myQueue.AddLast((Link*) myData);
                     HandleWrite(&tInfo->p, &myQueue);
                     break;
                  }
               }
            }
            break;
         }
      }
   }

   message = new TFTCPClientCommunicationPacket(0, TFTCPClientCommunicationPacket::TCPCLIENT_EVT_SHUTDOWN);
   message->shutDownThread = tInfo->p.hSocket;
   tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_TCPCLIENT_MANAGER, message);
   return(0);
}
/*ServerMonClient*/

/* -------------------------------------------------------------------------
   Name:        HandleRead
   ------------------------------------------------------------------------- */
bool HandleRead (MON_THREAD_INFO *tInfo, Head *myQueue, NET_INFO *info)
{
   int   bytesRcvd;
   int   err;
   bool  ret;

   // ---------------------------------------------------- READ IN THE NET_MSG.
   if(info->msgRemaining > 0)
   {
      bytesRcvd = recv (tInfo->p.hSocket, (char*)info->msgPtr,
                        info->msgRemaining, 0);
      if(bytesRcvd == SOCKET_ERROR)
      {
         err = WSAGetLastError();
         if (err == WSAEWOULDBLOCK)
            return true;   // This should never happen, but handle it anyway.
         else
            return false;  // Something went wrong, connection should close.
      }
      info->msgRemaining -= bytesRcvd;
      info->msgPtr        = ((char*)info->msgPtr) + bytesRcvd;
      if(info->msgRemaining == 0)
      {
         info->dataRemaining = info->msg.dataSize;
         if(info->msg.dataSize > 0)
         {
            info->clientData = new char [info->msg.dataSize];
            info->dataPtr = info->clientData;
            return true;   // Have NET_MSG, now wait for data.
         }
      }
      else
      {
         return true;      // Still waiting for more data for NET_MSG header.
      }
   }

   // ------------------------------------------------------- READ IN THE DATA.
   // If there is data attached to the NET_MSG structure,
   // wait for the data to arrive.  If the message does not
   // contain any data, skip over this section and process
   // the message.
   // -------------------------------------------------------
   if(info->dataRemaining > 0)
   {
      bytesRcvd = recv(tInfo->p.hSocket, (char*)info->dataPtr, info->dataRemaining, 0);
      if(bytesRcvd == SOCKET_ERROR)
      {
         err = WSAGetLastError();
         if(err == WSAEWOULDBLOCK)
         {
            return(true);   // This should never happen, but handle it anyway.
         }
         else
         {
            return false;  // Something went wrong, connection should close.
         }
      }
      info->dataRemaining -= bytesRcvd;
      info->dataPtr        = ((char*)info->dataPtr) + bytesRcvd;
      if(info->dataRemaining > 0)
      {
         return true;      // Still waiting for more data for NET_MSG data.
      }
   }

   // ---------------------------------------------------- PROCESS THE MESSAGE.
   switch(info->msg.id)
   {
      case NET_EVT_SEND_UPDATE:
      {
         ret = SendUpdate (tInfo, myQueue, (bool)info->msg.value);
         break;
      }

      case NET_EVT_SEND_OUTPUT:
      {
         ret = SendOutput (tInfo, myQueue, info);
         break;
      }

      case NET_EVT_SD_NORMAL:
      {
         ret = SendAppMessage (tInfo, info->msg.value, TFEngineCommunicationPacket::SRV_EVT_SD_NORMAL);
         break;
      }

      case NET_EVT_SD_ABORT:
      {
         ret = SendAppMessage (tInfo, info->msg.value, TFEngineCommunicationPacket::SRV_EVT_SD_ABORT);
         break;
      }

      case NET_EVT_RELOAD_NORMAL:
      {
         ret = SendAppMessage (tInfo, info->msg.value, TFEngineCommunicationPacket::SRV_EVT_RELOAD_NORMAL);
         break;
      }

      case NET_EVT_RELOAD_ABORT:
      {
         ret = SendAppMessage (tInfo, info->msg.value, TFEngineCommunicationPacket::SRV_EVT_RELOAD_ABORT);
         break;
      }

      case NET_EVT_ALL_SD_NORMAL:
      {
         ret = BroadcastAppMessage (tInfo, TFEngineCommunicationPacket::SRV_EVT_SD_NORMAL);
         break;
      }

      case NET_EVT_ALL_SD_ABORT:
      {
         ret = BroadcastAppMessage (tInfo, TFEngineCommunicationPacket::SRV_EVT_SD_ABORT);
         break;
      }

      case NET_EVT_ALL_RELOAD_NORMAL:
      {
         ret = BroadcastAppMessage(tInfo, TFEngineCommunicationPacket::SRV_EVT_RELOAD_NORMAL);
         break;
      }

      case NET_EVT_ALL_RELOAD_ABORT:
      {
         ret = BroadcastAppMessage(tInfo, TFEngineCommunicationPacket::SRV_EVT_RELOAD_ABORT);
         break;
      }

      case NET_EVT_RELOAD_TAL:
      {
         // Shut down threads
         ret = BroadcastAppMessage(tInfo, TFEngineCommunicationPacket::SRV_EVT_TERMINATE);

   		// Inform Server that we want to reload
	   	TFServerCommunicationPacket*	message = new TFServerCommunicationPacket(TFServerCommunicationPacket::SRV_EVT_RELOAD_TAL, 0);
         tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_SERVER, message);
         break;
      }

      case NET_EVT_START_STREAM:
      case NET_EVT_STOP_STREAM:
      case NET_EVT_PAUSE_STREAM:
      {
         IVRApp* app;
         int appID;

         tInfo->ivrApps->Lock();
         for(app = (IVRApp*) tInfo->ivrApps->First(); (app != NULL)
            && (app->data.iPort != info->msg.value); app = (IVRApp*) app->Suc());

// For debugging
// SOCKADDR_IN destination;
// int addressSize = sizeof(SOCKADDR_IN);
// getpeername(tInfo->p.hSocket, (sockaddr*) &destination, &addressSize);
// if(app != NULL)
// {
//    appID = app->data.iAppID;
// }
//
// if(info->msg.id == NET_EVT_START_STREAM)
// {
//    printf("Received start");
// }
// else if(info->msg.id == NET_EVT_STOP_STREAM)
// {
//    printf("Received stop");
// }
// else
// {
//    printf("Received pause");
// }
//
// printf(" from %s:%d for port %d", inet_ntoa(destination.sin_addr), ntohs(destination.sin_port), info->msg.value);
//
// if(app != NULL)
// {
//    printf(" (app %d)\n", appID);
// }
// else
// {
//    printf(" (No such app)\n");
// }

         if((app != NULL) && ((app->data.clientMsg.status == lsNormal) || (app->data.clientMsg.status == lsWarning)))
         {
            appID = app->data.iAppID;

            switch(info->msg.id)
            {
               case NET_EVT_START_STREAM: ret = SendAppStartStreamMessage(tInfo, appID, info->msg.value); break;
               case NET_EVT_STOP_STREAM: ret = SendAppStopStreamMessage(tInfo, appID); break;
               case NET_EVT_PAUSE_STREAM: ret = SendAppPauseStreamMessage(tInfo, appID); break;
            }
         }
         else
         {
            // App message didn't make it, app probably doesn't exist, so send ourselves a stop message
            TFTCPClientCommunicationPacket* stopMessage = new TFTCPClientCommunicationPacket(tInfo->p.hSocket, TFTCPClientCommunicationPacket::TCPCLIENT_EVT_STOP_STREAM);
            tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_TCPCLIENT_MANAGER, stopMessage);
         }

         tInfo->ivrApps->Unlock();
         break;
      }

      default:
      {
         ret = false;
      }
   }

   if(ret)
   {
      ret = HandleWrite(&tInfo->p, myQueue);
   }

   // Reset for next message.
   info->msgPtr = (void*) &(info->msg);
   info->msgRemaining = sizeof(NET_MSG);
   delete [] info->clientData;
   info->clientData = 0;
   info->dataPtr = 0;
   info->dataRemaining = 0;

   return ret;
}
/*HandleRead*/

/* -------------------------------------------------------------------------
   Name:        HandleWrite
   ------------------------------------------------------------------------- */
bool HandleWrite (NET_STORAGE *p, Head *myQueue)
{
   SendData  *myData;
   SendData  *tempData;
   int        bytesSent;
   int        err;

   myData = (SendData*) myQueue->First();

   while (myData) {
      bytesSent = send (p->hSocket, myData->sendPtr, myData->remaining, 0);
      if (bytesSent == SOCKET_ERROR) {
         err = WSAGetLastError();
         if (err == WSAEWOULDBLOCK)
            return true;   // Buffer full, wait for FD_WRITE.
         else
            return false;  // Something went wrong, connection should close.
      }
      myData->remaining -= bytesSent;
      myData->sendPtr   += bytesSent;
      if (myData->remaining == 0) {
         tempData = myData;
         myData = (SendData*) myData->Suc();
         delete tempData;
      }
   }
   return true;
}
/*HandleWrite*/

/* -------------------------------------------------------------------------
   Name:        FlagChangedApplications

   Purpose:     Updates the status of all applications in the application
                list flagging those applications that have changed since the
                last update.  If "bRefresh" is set, all applications are
                flagged as changed.
   ------------------------------------------------------------------------- */
bool FlagChangedApplications (
      MON_THREAD_INFO *tInfo,
      bool             bRefresh,
      int             &sendInfoCnt)
{
   IVRApp      *temp;
   int          msgNo;
   int          appCount;
   int          offset;
   DWORD        ret;

   sendInfoCnt = 0;
   if (tInfo->ivrApps->Lock()) {
      appCount = tInfo->ivrApps->Cardinal();
      // If the number of applications in the list change, re-allocate memory.
      if (appCount != tInfo->appInfoCnt) {
         delete [] tInfo->appInfo;
         tInfo->appInfoCnt = appCount;
         tInfo->appInfo = new EXPANDED_CLIENT_MSG [tInfo->appInfoCnt];
         memset (tInfo->appInfo, 0, sizeof (EXPANDED_CLIENT_MSG) * tInfo->appInfoCnt);
      }
      offset = 0;
      temp = (IVRApp*) tInfo->ivrApps->First();
      // Determine which applications have changed.
      // By only sending changed applications, the Monitor doesn't flicker.
      while (temp) {
         ret = WaitForSingleObject (temp->data.dataMutex, INFINITE);
         if (ret == WAIT_OBJECT_0) {
            msgNo = temp->data.clientMsg.msgNo;
            // Monitor may request a refresh.  If so, send current
            // state of all applications.
            if (bRefresh || msgNo != tInfo->appInfo[offset].msgNo) {
               memcpy (&(tInfo->appInfo[offset]),
                       &(temp->data.clientMsg),
                       sizeof (EXPANDED_CLIENT_MSG));
               SetEvent(tInfo->appInfo[offset].hChanged);
               sendInfoCnt++;
            }
            // Is hChanged set (i.e. TRUE)?
            else if(WaitForSingleObject(tInfo->appInfo[offset].hChanged, 0) == WAIT_OBJECT_0)
               sendInfoCnt++;
            offset++;
            ReleaseMutex (temp->data.dataMutex);
         }
         temp = (IVRApp*) temp->Suc();
      }
      tInfo->ivrApps->Unlock();
      return true;
   }
   return false;
}
/*FlagChangedApplications*/

/* -------------------------------------------------------------------------
   Name:        SendUpdate

   Purpose:     The Monitor is requesting an update in application status.
                If "bRefresh" is set, the status of all applications is sent
                to the Monitor.  If "bRefresh" is not set, only changes
                since the last update will be sent.
   ------------------------------------------------------------------------- */
bool SendUpdate (MON_THREAD_INFO *tInfo, Head *myQueue, bool bRefresh)
{
   EXPANDED_CLIENT_MSG  *sendInfo    = 0;
   SendData    *myData;
   int          sendInfoCnt;
   int          appOffset;
   int          offset;
   NET_MSG      msg;

   if (FlagChangedApplications (tInfo, bRefresh, sendInfoCnt)) {
      sendInfo = new EXPANDED_CLIENT_MSG [sendInfoCnt];
      appOffset = 0;
      offset = 0;
      while(appOffset < sendInfoCnt)
      {
         // Is hChanged set (i.e. TRUE)?
         if(WaitForSingleObject(tInfo->appInfo[offset].hChanged, 0) == WAIT_OBJECT_0)
         {
            memcpy (&(sendInfo[appOffset]),
                    &(tInfo->appInfo[offset]),
                    sizeof (EXPANDED_CLIENT_MSG));
            appOffset++;
            ResetEvent(tInfo->appInfo[offset].hChanged);
         }
         offset++;
      }
      msg.dataSize = sendInfoCnt * sizeof (EXPANDED_CLIENT_MSG);
   } else
      msg.dataSize = 0;

   msg.id = NET_EVT_UPDATE;
   msg.value = 0;
   msg.data = 0;
   myData = new SendData (&msg, sizeof (NET_MSG));
   myQueue->AddLast ((Link*)myData);
   if (msg.dataSize) {
      myData = new SendData (sendInfo, msg.dataSize);
      myQueue->AddLast ((Link*)myData);
   }
   delete [] sendInfo;

   return true;
}
/*SendUpdate*/

/* -------------------------------------------------------------------------
   Name:        SendOutput
   ------------------------------------------------------------------------- */
bool SendOutput (MON_THREAD_INFO *tInfo, Head *myQueue, NET_INFO *info)
{
   IVRApp   *temp;
   int       lineSize;
   int       lineCount;
   int       len = 0;
   DWORD     ret;
   char     *lines = 0;
   bool      bFound = false;
   SendData *myData;

   // Lock the application list.
   if (tInfo->ivrApps->Lock()) {
      // Locate the application.
      temp = (IVRApp*) tInfo->ivrApps->First();
      while (temp && !bFound) {
         if (temp->data.iAppID == info->msg.value)
            bFound = true;
         else
            temp = (IVRApp*) temp->Suc();
      }
      // If we found the application, retrieve the output lines.
      if (temp) {
         // Lock the application data.
         ret = WaitForSingleObject (temp->data.dataMutex, INFINITE);
         if (ret == WAIT_OBJECT_0) {
            // Retrieve the output lines.
            lineSize = temp->data.ivLog->GetMaxLineSize();
            lineCount = temp->data.ivLog->GetMaxLines();
            lines = new char [lineSize * lineCount + lineCount];
            len = temp->data.ivLog->GetLines (lines, INFINITE);

            // Unlock the application data.
            ReleaseMutex (temp->data.dataMutex);
         }
      }
      // Unlock the application list.
      tInfo->ivrApps->Unlock();
   }

   info->msg.id       = NET_EVT_OUTPUT;
// info->msg.value    = info->msg.value;
   info->msg.data     = 0;
   info->msg.dataSize = len;
   myData = new SendData (&(info->msg), sizeof (NET_MSG));
   myQueue->AddLast ((Link*)myData);
   if (len) {
      myData = new SendData (lines, len);
      myQueue->AddLast ((Link*)myData);
   }

   delete [] lines;
   return bFound;
}
/*SendOutput*/

/* -------------------------------------------------------------------------
   Name:        SendAppMessage
   ------------------------------------------------------------------------- */
bool SendAppMessage(MON_THREAD_INFO* tInfo, int appID, TFEngineCommunicationPacket::TF_ENGINE_EVENT eventID)
{
   TFEngineCommunicationPacket* message = new TFEngineCommunicationPacket(appID, eventID);
   tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, message);
   return(true);
}
/*SendAppMessage*/

/* -------------------------------------------------------------------------
   Name:        BroadcastAppMessage
   ------------------------------------------------------------------------- */
bool BroadcastAppMessage(MON_THREAD_INFO* tInfo, TFEngineCommunicationPacket::TF_ENGINE_EVENT eventID)
{
   TFEngineCommunicationPacket* message = new TFEngineCommunicationPacket(0, eventID);
   tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, message);
   return(true);
}
/*BroadcastAppMessage*/

bool SendAppStopStreamMessage(MON_THREAD_INFO* tInfo, int appID)
{
   TFEngineCommunicationPacket* message;
   int addressSize = sizeof(SOCKADDR_IN);

   if((appID > 1000) && (tInfo->appID != 0))
   {
      message = new TFEngineCommunicationPacket(appID, TFEngineCommunicationPacket::SRV_EVT_STOP_STREAM);
      getpeername(tInfo->p.hSocket, (sockaddr*) &(message->destination), &addressSize);
      message->destination.sin_port = htons(tInfo->internetPort);

      tInfo->appID = 0;
      
      tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, message);
   }

   return(true);
}

bool SendAppStartStreamMessage(MON_THREAD_INFO* tInfo, int appID, int telephonyPort)
{
   SOCKADDR_IN hostAddress;
   TFEngineCommunicationPacket* message;
   int addressSize = sizeof(SOCKADDR_IN);
   short internetPort;

   if(appID > 1000)
   {
      message = new TFEngineCommunicationPacket(appID, TFEngineCommunicationPacket::SRV_EVT_START_STREAM);
      message->threadID = tInfo->p.hSocket;
      getsockname(tInfo->p.hSocket, (sockaddr*) &hostAddress, &addressSize);
      getpeername(tInfo->p.hSocket, (sockaddr*) &(message->destination), &addressSize);
      internetPort = (short) (ntohs(hostAddress.sin_port) + telephonyPort);
      message->destination.sin_port = htons(internetPort);

      if(((tInfo->appID != appID) || (tInfo->internetPort != internetPort)) && (tInfo->appID != 0))
      {
         SendAppStopStreamMessage(tInfo, tInfo->appID);
      }

      tInfo->appID = appID;
      tInfo->internetPort = internetPort;
      tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, message);
   }

   return(true);
}

bool SendAppPauseStreamMessage(MON_THREAD_INFO* tInfo, int appID)
{
   TFEngineCommunicationPacket* message;
   int addressSize = sizeof(SOCKADDR_IN);

   if((appID > 1000) && (tInfo->appID != 0))
   {
      message = new TFEngineCommunicationPacket(appID, TFEngineCommunicationPacket::SRV_EVT_PAUSE_STREAM);
      message->threadID = tInfo->p.hSocket;
      getpeername(tInfo->p.hSocket, (sockaddr*) &(message->destination), &addressSize);
      message->destination.sin_port = htons(tInfo->internetPort);

      tInfo->commHub->SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, message);
   }

   return(true);
}

