/* -------------------------------------------------------------------------
   File:        SIGNALEVT.H

   Purpose:     This template class is used to send events between different
                threads.  It's defined as a template to alow different
                "events" to be sent.

   Notes:       - Templates don't get compiled like .CPP files.  The
                  functions must be defined in the .H file so they can be
                  included in other files.
   ------------------------------------------------------------------------- */
#ifndef SIGNALEVT_H
#define SIGNALEVT_H

#include <windows.h>

#include <deque>;
using std::deque;

#include "..\escutil\CriticalSection.h"

/* -------------------------------------------------------------------------
   Structure:   SRVR_EVT
   ------------------------------------------------------------------------- */
typedef struct
{
   int          event;
   HANDLE       hThread;
   unsigned (__stdcall *start)(void *);
} SRVR_EVT;

/* -------------------------------------------------------------------------
   Class:       SignalEvent

   Notes:       - It would be fairly easy to turn this into a template
                  class.
   ------------------------------------------------------------------------- */
template <class T> class SignalEvent
{
public:
   SignalEvent (unsigned queueSize = 32);
   ~SignalEvent();
   HANDLE GetHandle();
   bool AddEvent (T &event, DWORD timeout = INFINITE);
   bool GetEvent (T &event, DWORD timeout = INFINITE);
   bool PeekEvent(T &event, DWORD timeout = INFINITE);
   bool Empty(void) { return first == last; };
   bool Clear (DWORD timeout = INFINITE);
   int Size(void) { return(eventCount); };
private:
   HANDLE     hEvent;
   HANDLE     hMutex;
   T         *events;
   unsigned   first;
   unsigned   last;
   unsigned   eventCount;
   unsigned   maxCount;
};

/* -------------------------------------------------------------------------
   Name:        SignalEvent::SignalEvent

   Purpose:     Creates event queue and wait objects to synchronize access
                to queue and signal when events arrive.

   Notes:       - The minimum queue size is 5.
   ------------------------------------------------------------------------- */
template <class T> SignalEvent<T>::SignalEvent (unsigned queueSize)
{
   first = last = eventCount = 0;

   // Signalled when events are added to the queue.
   hEvent = CreateEvent (0, false, false, 0);

   // Must acquire mutex before adding events to the queue.
   hMutex = CreateMutex (0, false, 0);

   if (queueSize < 5)
      queueSize = 5;
   maxCount = queueSize;

   events = new T [maxCount];
}
/*SignalEvent::SignalEvent*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::~SignalEvent

   Purpose:     Deletes the event queue and wait objects.
   ------------------------------------------------------------------------- */
template <class T> SignalEvent<T>::~SignalEvent()
{
   delete [] events;
   CloseHandle (hMutex);
   CloseHandle (hEvent);
}
/*SignalEvent::~SignalEvent*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::GetHandle

   Purpose:     The event handle returned from this function is signalled
                when events have been placed on the queue.
   ------------------------------------------------------------------------- */
template <class T> HANDLE SignalEvent<T>::GetHandle()
{
   return hEvent;
}
/*SignalEvent::GetHandle*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::Clear
   ------------------------------------------------------------------------- */
template <class T> bool SignalEvent<T>::Clear (DWORD timeout)
{
   DWORD  ret;

   ret = WaitForSingleObject (hMutex, timeout);
   if (ret != WAIT_OBJECT_0)
      return false;

   first = last = eventCount = 0;

   ResetEvent(hEvent);

   ReleaseMutex (hMutex);
   return true;
}
/*SignalEvent::Clear*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::AddEvent

   Purpose:     Attempts to add an event to the queue and signal a listening
                thread that an event is waiting.  If the queue is full, this
                function fails.

   Returns:     true  - Successfully added an event to the queue.
                false - Timeout occurred or queue is full.
   ------------------------------------------------------------------------- */
template <class T> bool SignalEvent<T>::AddEvent (T &event, DWORD timeout)
{
   DWORD  ret;

   ret = WaitForSingleObject (hMutex, timeout);
   if(ret != WAIT_OBJECT_0)
   {
      return(false);
   }

   if(eventCount >= maxCount)
   {
      ReleaseMutex(hMutex);
      return(false);
   }

   memcpy (&events[last], &event, sizeof(T));
   last++;
   if (last == maxCount)
      last = 0;

   eventCount++;
   SetEvent (hEvent);
   ReleaseMutex (hMutex);
   return true;
}
/*SignalEvent::AddEvent*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::GetEvent

   Purpose:     Attempts to retrieve an event from the queue.  If the queue
                is

   Returns:     true  - Successfully retrieved an event from the queue.
                false - Timeout occurred or queue is empty.
   ------------------------------------------------------------------------- */
template <class T> bool SignalEvent<T>::GetEvent (T &event, DWORD timeout)
{
   DWORD  ret;

   ret = WaitForSingleObject (hMutex, timeout);
   if (ret != WAIT_OBJECT_0)
      return false;

   if (eventCount == 0) {
      ReleaseMutex (hMutex);
      return false;
   }

   memcpy (&event, &events[first], sizeof(T));
   first++;
   if (first == maxCount)
      first = 0;

   eventCount--;

	if(eventCount)
		SetEvent (hEvent);

   ReleaseMutex(hMutex);
   return true;
}
/*SignalEvent::GetEvent*/

/* -------------------------------------------------------------------------
   Name:        SignalEvent::PeekEvent

   Purpose:     Copies the event at the front of the queue

   Returns:     true  - Successfully copied the event at the front of the queue.
                false - The queue is empty.
   ------------------------------------------------------------------------- */
template <class T> bool SignalEvent<T>::PeekEvent(T &event, DWORD timeout)
{
   if(WaitForSingleObject(hMutex, timeout) != WAIT_OBJECT_0)
   {
      return(false);
   }

	if(eventCount != 0)
   {
	   memcpy(&event, &events[first], sizeof(T));
   }

   ReleaseMutex(hMutex);

   return(eventCount != 0);
}

// Designed as a replacement for SignalEvent<T>
// Dynamic queue length
// Queue may hold items of any type with an assignment constructor
// Thread safe
template <class T> class SignalQueue
{
   public:

      SignalQueue(void)
      {
         itemsMutex.Get();
         itemsInQueue = CreateEvent(NULL, false, false, NULL);
         itemsMutex.Release();
      };

      ~SignalQueue(void)
      {
         itemsMutex.Get();
         CloseHandle(itemsInQueue);
         itemsMutex.Release();
      };

      // Makes a copy of the item and adds it to the queue
      // Sets signal to indicate elements in queue
      void Add(T item)
      {
         itemsMutex.Get();
         items.push_back(item);
         SetEvent(itemsInQueue);
         itemsMutex.Release();
      };

      // Removes all elements from the queue
      // Resets signal to indicate no elements in queue
      void Clear(void)
      {
         itemsMutex.Get();
         ResetEvent(itemsInQueue);
         items.clear();
         itemsMutex.Release();
      };

      // Returns true if elements in queue, returns false otherwise
      bool Empty(void)
      {
         return(items.size() == 0);
      };

      // Sets item by reference to first element in queue
      // Removes element from queue
      // Resets signal if there are no elements in queue
      // Returns true if there was an element in the queue, otherwise returns false
      // and item is not changed
      bool Get(T& item)
      {
         itemsMutex.Get();

         if(!Peek(item))
         {
            itemsMutex.Release();
            return(false);
         }

         items.pop_front();
         if(items.size() != 0)
         {
            SetEvent(itemsInQueue);
         }
         itemsMutex.Release();
         return(true);
      };

      // Returns signal that is set if there are elements in queue
      HANDLE GetSignal(void)
      {
         return(itemsInQueue);
      };

      // Sets item by reference to first element in queue
      // Returns true if there was an element in the queue, otherwise returns false
      // and item is not changed
      bool Peek(T& item)
      {
         itemsMutex.Get();
         if(items.size() == 0)
         {
            itemsMutex.Release();
            return(false);
         }

         item = items.front();
         itemsMutex.Release();
         return(true);
      };

      // Returns the number of elements in queue
      int Size(void)
      {
         return(items.size());
      };

   protected:

      Mutex itemsMutex;
      HANDLE itemsInQueue;
      deque<T> items;
};

#endif
