#ifndef VARCONTROLLER_H
#define VARCONTROLLER_H

#include "..\TFServer\gui_log.h"
class BoardGeneric;

#include <string>
using std::string;
#include <map>
using std::map;
#include <stack>
using std::stack;

typedef map<string, string> VariableList;
typedef stack<VariableList> VariableListStack;

#if defined (__DLL__)
#define VARCONTAPI __declspec(dllexport)
#else
#define VARCONTAPI __declspec(dllimport)
#endif

class VARCONTAPI VarController
{
   public:

      // The VarController is initialised with a pointer to the engine log, this
      // will become a problem if the log becomes dynamic
      VarController(GUILog* ivLog, PAPP_STATUS_FLAGS pAppInfo);

      // Removes all variables including pushed local variables
      void Clear(void);

      // Returns a pointer to a newly allocated string of the complete path for
      // a given filename using the @SYS_DRIVE and @SYS_DIR system variables as necessary
      // Returns NULL and generates error message in log on error
      char* FilePath(const char* filename);

      // Searches for a given filename using the given pathname.
      // Returns a pointer to a newly allocated string of the complete path for
      // the given filename where it is first found, using our search
      // criteria, as follows:
      // (1) if filename contains a path specification (any at all),
      //     this method uses FilePath() instead to evaluate the
      //     given filename relative to the .TAP file location
      // (2) if pathname is NULL or empty, FilePath() is called instead,
      //     allowing legacy performance to be preserved
      // (3) check each folder location in the 'pathname' parameter, starting
      //     at the beginning, until file is found or end of path is reached
      // (4) if file not found in path, FilePath() is used to check for the
      //     file in the TAP location folder.
      // (5) if the file is not found in path, this function will return
      //     the filename prepended with the TAP path; see FilePath()
      // Notes:
      //  -  Pathname must be a valid Windows path expression, example:
      //        "C:\Program Files\TeleFlow;C:\MyAPP;C:\MyAPP\Vox\Eng\Scripts"
      //  -  If the pathname is invalid, the performance is unpredictable,
      //     but will likey result in the file not being found
      char* FilePathSearch(const char* filename, const char* pathname);

      // Returns a pointer to a constant string of the value of variable name
      // Returns NULL and generates error message in log on error
      const char* Get(const char* name);

      // Returns true if pszString is NULL or ""
      bool IsEmpty(const char* pszString);

      // Returns true if pszString evaluates to NULL (some variables undefined)
      // or "" with all variables replaced by their values
      bool IsParsedEmpty(const char* pszString);

      // Returns true if all variables in pszString are defined
      bool IsParsedInvalid(const char* pszString);

      // Returns true if name is a system variable
      bool IsSystem(const char* name);

      // Returns true if name is a defined variable
      bool IsVariable(const char* name);

      // Returns true if pszString is of valid variable format: "@VARNAME"
      bool IsVariableFormat(const char* pszString);

      // Returns IVR_SUCCESS if able to load the contents of file fileName into variable varName
      // Optionally convert file contents to Base64 before storing in varName.
      // Optionally count the number of bytes loaded or converted (if bToBase64 is true)
      // If converting to Base64, the count may include/exclude the added CR/LF chars (bCountCRLF)
      // Counting CR/LF only occurs if bToBase64 is true
      // Generates error message in log on error; returns IVR_FAIL if unable to find
      // or open the file, returns IVR_FATAL on errors (eg: bad variable name)
      IV_STATUS LoadFile(const char* varName, const char* fileName,
         unsigned int &iInputCount, unsigned int &iFinalCount,
         bool bToBase64 = false, bool bBreakLines = false, bool bCountCRLF = false);

      // Write contents of variable varName to file fileName, and returns IVR_SUCCESS.
      // Optionally convert contents of varName from Base64 to binary before writing.
      // Generates error message in log on error; returns IVR_FAIL if unable to write, or
      // returns IVR_FATAL on fatal errors (eg: variable not defined)
      IV_STATUS WriteFile(const char* fileName, const char* varName, bool bFromBase64 = false);

      // Returns a pointer to a newly allocated string of the value of pszString
      // with all variables replaced by their values
      // Returns NULL and generates error message in log on error
      char* Parse(const char* pszString);

      // Case 1a: parameterValue parses to itself and is non-empty (no change)
      //          Returns a pointer to a newly allocated string of parsed parameterValue
      //          Prints ("      %s is '%s'\n", parameterName, parsedValue) to log
      // Case 1b: parameterValue parses to a different non-empty value (change)
      //          Returns a pointer to a newly allocated string of parsed parameterValue
      //          Prints ("      %s '%s' evaluates to '%s'\n", parameterName, parameterValue, parsedValue) to log
      // Case 2:  parameterValue parses to NULL result (error)
      //          Returns a NULL pointer
      //          Prints ("      ERROR: %s '%s' is undefined\n", parameterName, parameterValue) to log
      // Case 3a: parameterValue parses to empty and defaultValue is NULL
      //          Returns a NULL pointer
      //          Prints ("      ERROR: %s cannot evaluate empty\n", parameterName) to log
      // Case 3b: parameterValue parses to empty, defaultValue is not NULL, defaultMessage is NULL,
      //          and defaultValue parses to itself (no change)
      //          Returns a pointer to a newly allocated string of parsed defaultValue
      //          Prints ("      Empty %s defaults to '%s'\n", parameterName, parsedValue) to log
      // Case 3c: parameterValue parses to empty, defaultValue is not NULL, defaultMessage is NULL,
      //          and defaultValue parses to a different non-empty value (change)
      //          Returns a pointer to a newly allocated string of parsed defaultValue
      //          Prints ("      Empty %s defaults to '%s', and evaluates to '%s'\n", parameterName, defaultValue, parsedValue) to log
      // Case 3d: parameterValue parses to empty, defaultValue is not NULL, defaultMessage is not NULL
      //          Returns a pointer to a newly allocated string of parsed defaultValue
      //          Prints ("      %s\n", defaultMessage) to log
      char* ParseParameter(const char* parameterName, const char* parameterValue,
         const char* defaultValue = NULL, const char* defaultMessage = NULL);

      // Used for passing values into a new TAM
      // Returns a pointer to a newly allocated string of the value of pszString
      // with all variables replaced by their values where local variable values
      // come from the previous TAM level
      // Returns NULL and generates error message in log on error
      char* ParseWithPushedVariables(const char* pszString);

      // Resets the local variables to their values in the previous TAM level
      // Generates error message in log on error
      void PopLocalVariables(void);

      // Used for synchronisation (e.g. COM)
      // Allows immediate processing of variables set with SetPending and processes
      // any currently pending sets
      // Generates error message in log on error
      void ProcessPendingOn(void);

      // Used for synchronisation (e.g. COM)
      // Sets processing of variables set with SetPending to wait until ProcessPendingOn
      void ProcessPendingOff(void);

      // Stores current local variables and creates a new blank list of local
      // variables for a new level of TAM
      void PushLocalVariables(void);

      // Returns true if able to set variable name to value with scope modifier
      // If the scope is not specified, DEFAULT will set local variable name to
      // value if it exists, otherwise will set global variable name to value
      // Cannot set system variables
      // Generates error message in log if not able
      bool Set(const char* name, const char* value, bool bSuppressPrint = false);

      // Sets ivBoard property to BoardGeneric to enable certain system variables:
      // @CALL_ELAPSE, @CALL_END, @CALL_SEIZED, @CALL_START
      void SetBoard(BoardGeneric* board);

      // Returns true if able to set global variable name to value
      // Generates error message in log if not able
      bool SetGlobal(const char* name, const char* value, bool bSuppressPrint = false);

      // Returns true if able to set local variable name to value
      // Generates error message in log if not able
      bool SetLocal(const char* name, const char* value, bool bSuppressPrint = false);

      // Used for synchronisation (e.g. COM)
      // Returns true if ProcessPendingOff or ProcessPendingOn AND able to set
      // variable name to value
      // Will set local variable name to value if it exists, otherwise will set
      // global variable name to value
      // Cannot set system variables
      // Generates error message in log if not able
      bool SetPending(const char* name, const char* value);

      // Used for returning reference values from a TAM
      // Returns true if able to set variable name of previous level of TAM to value
      // Generates error message in log if not able
      bool SetPopVariable(const char* name, const char* value);

      // Returns true if able to set system variable name to value
      // Generates error message in log if not able
      bool SetSystem(const char* name, const char* value);

      // Convert the source buffer into Base64 (MIME).
      // Create a buffer to store results in and return pointer to that buffer
      // Return value of NULL indicates an error.
      // Calling method is responsible for deletion.
      char* Base64Encode(const char* pSource, size_t nBufferLen,
         size_t &nConvertedCount, bool bBreakLines, bool bCountCRLF);

      // Convert the source buffer from Base64 (MIME) into its original form, and
      // Create a buffer to store results in and return pointer to that buffer.
      // Return value of NULL indicates an error.
      // Calling method is responsible for deletion.
      char* Base64Decode(const char* pSource, size_t nBufferLen, size_t &nReturnLen);

   private:

      VariableList globalVariables;            // List of global variables
      VariableList systemVariables;            // List of system variables
      VariableListStack localVariablesStack;    // Stack of local variable lists
      VariableList pendingVariables;           // List of pending variables
      bool processPending;                 // true for ProcessPendingOn, false for ProcessPendingOff
      GUILog* ivLog;                       // pointer to the engine log
      BoardGeneric* ivBoard;               // pointer to the engine board
      PAPP_STATUS_FLAGS  pAppFlags;     // pointer to status flags in Thread application information
};

#undef VARCONTAPI

#endif
